# Written by Konrad Hinsen <hinsen@cnrs-orleans.fr>
# with contributions from Greg Ward

"""
Physical quantities with units.

This module derives from `Konrad Hinsen`_'s :class:`PhysicalQuantity`
<http://dirac.cnrs-orleans.fr/ScientificPython/>.

This module provides a data type that represents a physical
quantity together with its unit. It is possible to add and
subtract these quantities if the units are compatible, and
a quantity can be converted to another compatible unit.
Multiplication, subtraction, and raising to integer powers
is allowed without restriction, and the result will have
the correct unit. A quantity can be raised to a non-integer
power only if the result can be represented by integer powers
of the base units.

The values of physical constants are taken from the 2002
recommended values from CODATA_. Other conversion factors
(e.g. for British units) come from `Appendix B of NIST Special Publication 811`_.

.. warning::

   We can't guarantee for the correctness of all entries in the unit table,
   so use this at your own risk!

.. _Konrad Hinsen:                              mailto:hinsen@cnrs-orleans.fr
.. _CODATA:                                     http://www.codata.org/
.. _Appendix B of NIST Special Publication 811: http://physics.nist.gov/Pubs/SP811/appenB9.html
"""
from __future__ import division
from __future__ import unicode_literals
from builtins import object
from builtins import range
from builtins import map
from builtins import str
from future.utils import string_types
__docformat__ = 'restructuredtext'

import re

from fipy.tools import numerix
from fipy.tools.numerix import MA
from fipy.tools.numerix import umath

from fipy.tools.dimensions.NumberDict import _NumberDict
from functools import reduce

__all__ = ["PhysicalField", "PhysicalUnit"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

# Class definitions

class PhysicalField(object):
    """
    Physical field or quantity with units
    """


    def __init__(self, value, unit = None, array = None):
        """
        Physical Fields can be constructed in one of two ways:

          - `PhysicalField(*value*, *unit*)`, where `*value*` is a number of
            arbitrary type and `*unit*` is a string containing the unit name

                >>> print(PhysicalField(value = 10., unit = 'm'))
                10.0 m

          - `PhysicalField(*string*)`, where `*string*` contains both the value
            and the unit. This form is provided to make interactive use more
            convenient

                >>> print(PhysicalField(value = "10. m"))
                10.0 m


        Dimensionless quantities, with a `unit` of 1, can be specified
        in several ways

            >>> print(PhysicalField(value = "1"))
            1.0 1
            >>> print(PhysicalField(value = 2., unit = " "))
            2.0 1
            >>> print(PhysicalField(value = 2.))
            2.0 1

        Physical arrays are also possible (and are the reason this code
        was adapted from `Konrad Hinsen`_'s original PhysicalQuantity_).
        The `value` can be a Numeric_ `array`:

            >>> a = numerix.array(((3., 4.), (5., 6.)))
            >>> print(PhysicalField(value = a, unit = "m"))
            [[ 3.  4.]
             [ 5.  6.]] m

        or a `tuple`:

            >>> print(PhysicalField(value = ((3., 4.), (5., 6.)), unit = "m"))
            [[ 3.  4.]
             [ 5.  6.]] m

        or as a single value to be applied to every element of a supplied array:

            >>> print(PhysicalField(value = 2., unit = "m", array = a))
            [[ 2.  2.]
             [ 2.  2.]] m

        Every element in an array has the same unit, which is stored only
        once for the whole array.

        .. _Konrad Hinsen: mailto:hinsen@cnrs-orleans.fr
        .. _PhysicalQuantity: http://dirac.cnrs-orleans.fr/ScientificPython/
        .. _Numeric: http://www.numpy.org
        """
        if hasattr(value, "value") and not isinstance(value, PhysicalField):
            value = value.value
        elif hasattr(value, "getValue") and callable(value.getValue):
            value = value.getValue()

        if isinstance(value, PhysicalField):
            if unit is not None:
                if not value.unit.isDimensionless():
                    value = value.inUnitsOf(unit)
                unit = _findUnit(unit)
            else:
                unit = value.unit

            if hasattr(value.value, 'copy'):
                value = value.value.copy()
            else:
                value = value.value
        elif unit is not None:
            unit = _findUnit(unit)
        elif isinstance(value, string_types):
            s = value.strip()
            match = PhysicalField._number.match(s)
            if match is None:
                value = 1
                unit = _findUnit(s)
#                   raise TypeError, 'No number found'
            else:
                value = float(match.group(0))
                unit = _findUnit(s[len(match.group(0)):])

        if type(value) in [type([]), type(())]:
            value = [PhysicalField(item, unit) for item in value]
            if unit is None:
                unit = value[0].unit
            normalized = []
            for item in value:
                if item.unit == unit:
                    normalized += [item.value]
                else:
                    normalized += [item.inUnitsOf(unit).value]
            value = numerix.array(normalized)

        if unit is None:
            unit = _unity
##             unit = _findUnit("")

        self.value = numerix.array(value)
        self._unit = unit
        if array is not None:
            array[:] = self.value
            self.value = array

    _number = re.compile('[+-]?[0-9]+(\\.[0-9]*)?([eE][+-]?[0-9]+)?')

    def ravel(self):
        return self.value.ravel()

    def copy(self):
        """
        Make a duplicate.

            >>> a = PhysicalField(1, unit = 'inch')
            >>> b = a.copy()

        The duplicate will not reflect changes made to the original

            >>> a.convertToUnit('cm')
            >>> print(a)
            2.54 cm
            >>> print(b)
            1 inch

        Likewise for arrays

            >>> a = PhysicalField(numerix.array((0, 1, 2)), unit  = 'm')
            >>> b = a.copy()
            >>> a[0] = 3
            >>> print(a)
            [3 1 2] m
            >>> print(b)
            [0 1 2] m

        """
        if hasattr(self.value, 'copy'):
            return PhysicalField(value = self.value.copy(), unit = self.unit)
        else:
            return PhysicalField(value = self.value, unit = self.unit)

    def __str__(self):
        """
        Return human-readable form of a physical quantity

            >>> print(PhysicalField(value = 3., unit = "eV"))
            3.0 eV
        """
        return str(self.value) + ' ' + self.unit.name()

    def __repr__(self):
        """
        Return representation of a physical quantity suitable for re-use

            >>> PhysicalField(value = 3., unit = "eV")
            PhysicalField(3.0,'eV')
        """
        value = self.value
        if value.shape == ():
            value = str(value)
        else:
            value = repr(value)

        if self.unit is _unity:
            return value
        else:
            return (self.__class__.__name__ + '(' + value + ',' +
                    repr(self.unit.name()) + ')')

    def tostring(self, max_line_width=75, precision=8, suppress_small=False, separator=' '):
        """
        Return human-readable form of a physical quantity

            >>> p = PhysicalField(value = (3., 3.14159), unit = "eV")
            >>> print(p.tostring(precision = 3, separator = '|'))
            [ 3.   | 3.142] eV
        """
        from fipy.tools import numerix
        return numerix.tostring(self.value, max_line_width = max_line_width,
                                precision = precision,
                                suppress_small = suppress_small,
                                separator = separator) + ' ' + self.unit.name()


    def _sum(self, other, sign1 = lambda a: a, sign2 = lambda b: b):
        selfValue = self.value

        if _isVariable(other):
            return sign2(other) + self.__class__(value = sign1(selfValue), unit = self.unit)

        if isinstance(other, string_types):
            other = PhysicalField(value = other)

        if not isinstance(other, PhysicalField):
            if numerix.alltrue(other == 0):
                new_value = sign1(selfValue)
            elif self.unit.isDimensionlessOrAngle() or self.unit.isInverseAngle():
                new_value = sign1(selfValue) + sign2(other)
            else:
                raise TypeError(str(self) + ' and ' + str(other) + ' are incompatible.')
        else:
            new_value = (sign1(selfValue)
                         + sign2(other.value) * other.unit.conversionFactorTo(self.unit))
        return self.__class__(value=new_value, unit=self.unit)

    def __add__(self, other):
        """
        Add two physical quantities, so long as their units are compatible.
        The unit of the result is the unit of the first operand.

            >>> print(PhysicalField(10., 'km') + PhysicalField(10., 'm'))
            10.01 km
            >>> print(PhysicalField(10., 'km') + PhysicalField(10., 'J'))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return self._sum(other)

    __radd__ = __add__
    add = __add__

    def __sub__(self, other):
        """
        Subtract two physical quantities, so long as their units are compatible.
        The unit of the result is the unit of the first operand.

            >>> print(PhysicalField(10., 'km') - PhysicalField(10., 'm'))
            9.99 km
            >>> print(PhysicalField(10., 'km') - PhysicalField(10., 'J'))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return self._sum(other, sign2 = lambda b: -b)

    subtract = __sub__

    def __rsub__(self, other):
        return self._sum(other, sign1 = lambda a: -a)

    def __mul__(self, other):
        """
        Multiply two physical quantities.  The unit of the result is the
        product of the units of the operands.

            >>> print(PhysicalField(10., 'N') * PhysicalField(10., 'm') == PhysicalField(100., 'N*m'))
            True

        As a special case, if the result is dimensionless, the value
        is returned without units, rather than with a dimensionless unit
        of `1`. This facilitates passing physical quantities to packages
        such as Numeric that cannot use units, while ensuring the quantities
        have the desired units.

            >>> print((PhysicalField(10., 's') * PhysicalField(2., 'Hz')))
            20.0
        """
        if _isVariable(other):
            return other.__mul__(self)
        if isinstance(other, string_types):
            other = PhysicalField(value = other)
        if not isinstance(other, PhysicalField):
            return self.__class__(value = self.value*other, unit = self.unit)
        value = self.value*other.value
        unit = self.unit*other.unit
        if unit.isDimensionless():
            if unit.factor != 1:
                return value * unit.factor
            else:
                return value
        else:
            return self.__class__(value = value, unit = unit)

    __rmul__ = __mul__
    multiply = __mul__

    def __truediv__(self, other):
        """
        Divide two physical quantities.  The unit of the result is the
        unit of the first operand divided by the unit of the second.

            >>> print(PhysicalField(10., 'm') / PhysicalField(2., 's'))
            5.0 m/s

        As a special case, if the result is dimensionless, the value
        is returned without units, rather than with a dimensionless unit
        of `1`. This facilitates passing physical quantities to packages
        such as Numeric_ that cannot use units, while ensuring the quantities
        have the desired units

            >>> print((PhysicalField(1., 'inch')
            ...        / PhysicalField(1., 'mm')))
            25.4

        .. _Numeric: http://www.numpy.org
        """
        if _isVariable(other):
            return other.__rtruediv__(self)
        if isinstance(other, string_types):
            other = self.__class__(value = other)
        if not isinstance(other, PhysicalField):
            value = self.value / other
            unit = self.unit
        else:
            value = self.value / other.value
            unit = self.unit / other.unit
        if unit.isDimensionless():
            return value*unit.factor
        else:
            return self.__class__(value = value, unit = unit)

    __div__ = __truediv__
    divide = __truediv__

    def __rtruediv__(self, other):
        if _isVariable(other):
            return other.__truediv__(self)
        if isinstance(other, string_types):
            other = PhysicalField(value = other)
        if not isinstance(other, PhysicalField):
            value = other / self.value
            unit = pow(self.unit, -1)
        else:
            value = other.value / self.value
            unit = other.unit / self.unit
        if unit.isDimensionless():
            return value*unit.factor
        else:
            return self.__class__(value = value, unit = unit)

    __rdiv__ = __rtruediv__

    def __mod__(self, other):
        """
        Return the remainder of dividing two physical quantities.  The unit of the result is the
        unit of the first operand divided by the unit of the second.

            >>> print(PhysicalField(11., 'm') % PhysicalField(2., 's'))
            1.0 m/s
        """
        if _isVariable(other):
            return other.__rmod__(self)
        if isinstance(other, string_types):
            other = self.__class__(value = other)
        if not isinstance(other, PhysicalField):
            value = self.value % other
            unit = self.unit
        else:
            value = self.value % other.value
            unit = self.unit / other.unit
        if unit.isDimensionless():
            return value*unit.factor
        else:
            return self.__class__(value = value, unit = unit)

    def __pow__(self, other):
        """
        Raise a `PhysicalField` to a power. The unit is raised to the same power.

            >>> print(PhysicalField(10., 'm')**2)
            100.0 m**2
        """
        if isinstance(other, string_types):
            other = PhysicalField(value = other)
        return self.__class__(value = pow(self.value, float(other)), unit = pow(self.unit, other))

    def __rpow__(self, other):
        return pow(other, float(self))

    def __abs__(self):
        """
        Return the absolute value of the quantity. The `unit` is unchanged.

            >>> print(abs(PhysicalField(((3., -2.), (-1., 4.)), 'm')))
            [[ 3.  2.]
             [ 1.  4.]] m
        """
        return self.__class__(value = abs(self.value), unit = self.unit)

    def __pos__(self):
        return self

    def __neg__(self):
        """
        Return the negative of the quantity. The `unit` is unchanged.

            >>> print(-PhysicalField(((3., -2.), (-1., 4.)), 'm'))
            [[-3.  2.]
             [ 1. -4.]] m
        """
        return self.__class__(value = -self.value, unit = self.unit)

    def sign(self):
        """
        Return the sign of the quantity. The `unit` is unchanged.

            >>> from fipy.tools.numerix import sign
            >>> print(sign(PhysicalField(((3., -2.), (-1., 4.)), 'm')))
            [[ 1. -1.]
             [-1.  1.]]
        """
        return numerix.sign(self.value)

    def __bool__(self):
        """
        Test if the quantity is zero.

        Should this only pass if the unit offset is zero?
        """
        return self.value != 0

    __nonzero__ = __bool__

    def _inMyUnits(self, other):
        if _isVariable(other):
            other = other.value

        if not isinstance(other, PhysicalField):
            if isinstance(other, string_types):
                other = PhysicalField(other)
            elif numerix.alltrue(other == 0) or self.unit.isDimensionlessOrAngle():
                other = PhysicalField(value = other, unit = self.unit)
            else:
                raise TypeError('Incompatible units')
        return other.inUnitsOf(self.unit)

    def __getitem__(self, index):
        """
        Return the specified element of the array. The unit of the result
        will be the unit of the array.

            >>> a = PhysicalField(((3., 4.), (5., 6.)), "m")
            >>> print(a[1, 1])
            6.0 m
        """
        return PhysicalField(self.value[index], self.unit)

    def __setitem__(self, index, value):
        """
        Assign the specified element of the array, performing appropriate
        conversions.

            >>> a = PhysicalField(((3., 4.), (5., 6.)), "m")
            >>> a[0, 1] = PhysicalField("6 ft")
            >>> print(a)
            [[ 3.      1.8288]
             [ 5.      6.    ]] m
            >>> a[1, 0] = PhysicalField("2 min")
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        if isinstance(value, string_types):
            value = PhysicalField(value)
        if isinstance(value, PhysicalField) or _isVariable(value):
            value = self._inMyUnits(value).value
        self.value[index] = value

    def itemset(self, value):
        """
        Assign the value of a scalar array, performing appropriate
        conversions.

            >>> a = PhysicalField(4., "m")
            >>> a.itemset(PhysicalField("6 ft"))
            >>> print(a.allclose("1.8288 m"))
            1
            >>> a = PhysicalField(((3., 4.), (5., 6.)), "m")
            >>> try: #doctest: +IGNORE_EXCEPTION_DETAIL
            ...     a.itemset(PhysicalField("6 ft"))
            ... except IndexError:
            ...     # NumPy 1.7 has changed the exception type
            ...     raise ValueError("can only place a scalar for an  array of size 1")
            Traceback (most recent call last):
                ...
            ValueError: can only convert an array of size 1 to a Python scalar
            >>> a.itemset(PhysicalField("2 min"))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        if isinstance(value, string_types):
            value = PhysicalField(value)
        if isinstance(value, PhysicalField) or _isVariable(value):
            value = self._inMyUnits(value).value
        self.value.itemset(value)


##    __array_priority__ and __array_wrap__ are required to override
##    the default behavior of numpy. If a numpy array and a Variable
##    are in a binary operation and numpy is first, then numpy will,
##    by default, try and do everything it can to get a a raw numpy
##    array out of Variable. __array_wrap__ seems to have been
##    introduced into masked array to fix this issue. __array_wrap__ is
##    called after the operation is done so it could hurt efficiency badly.
##    Something else needs to be done to stop the initial evaluation.

    __array_priority__ = 100.0

    def __array_wrap__(self, arr, context=None):
        """
        Required to prevent numpy not calling the reverse binary operations.
        Both the following tests are examples ufuncs.

        >>> from fipy.tools.dimensions.physicalField import PhysicalField
        >>> print(type(numerix.array([1.0, 2.0]) * PhysicalField([1.0, 2.0], unit="m")))
        <class 'fipy.tools.dimensions.physicalField.PhysicalField'>

        >>> print(type(numerix.array([1.0, 2.0]) * PhysicalField([1.0, 2.0])))
        <class 'fipy.tools.dimensions.physicalField.PhysicalField'>

        >>> from scipy.special import gamma as Gamma # doctest: +SCIPY
        >>> print(isinstance(Gamma(PhysicalField([1.0, 2.0])), type(numerix.array(1)))) # doctest: +SCIPY
        1
        """
        result = arr

        if context is not None:
            (func, args, _) = context
            def __makePhysical(v):
                if not (isinstance(v, PhysicalField) or _isVariable(v)):
                    v = PhysicalField(v)
                return v
            args = [__makePhysical(arg) for arg in args]

            meth = getattr(args[0], func.__name__, None)
            if meth is not None and callable(meth):
                result = meth(*args[1:])

        return result

    def __array__(self, t = None):
        """
        Return a dimensionless `PhysicalField` as a Numeric_ ``array``.

            >>> print(numerix.array(PhysicalField(((2., 3.), (4., 5.)), "m/m")))
            [[ 2.  3.]
             [ 4.  5.]]

        As a special case, fields with angular units are converted to base
        units (radians) and then assumed dimensionless.

            >>> print(numerix.array(PhysicalField(((2., 3.), (4., 5.)), "deg")))
            [[ 0.03490659  0.05235988]
             [ 0.06981317  0.08726646]]


        If the array is not dimensionless, the numerical value in its base
        units is returned.

            >>> numerix.array(PhysicalField(((2., 3.), (4., 5.)), "mm"))
            array([[ 0.002,  0.003],
                   [ 0.004,  0.005]])

        .. _Numeric: http://www.numpy.org
        """
        if self.unit.isAngle():
            value = self.inRadians()
        else:
            value = self.inSIUnits().value

        return numerix.array(value, t)

#         if self.unit.isDimensionlessOrAngle():
#             value = self.numericValue
# ##            if type(value) is type(numerix.array((0))) and (t is None or t == value.typecode()):
#             if type(value) is type(numerix.array((0))) and (t is None or t == value.dtype.char):
#                 return value
#             else:
#                 return numerix.array(self.numericValue, t)
#         else:
#             raise TypeError, 'Numeric array value must be dimensionless'

    @property
    def _array(self):
        if self.unit.isDimensionlessOrAngle():
            return self.value
        else:
            raise TypeError('Numeric array value must be dimensionless')

    def __float__(self):
        """
        Return a dimensionless `PhysicalField` quantity as a float.

            >>> float(PhysicalField("2. m/m"))
            2.0

        As a special case, quantities with angular units are converted to
        base units (radians) and then assumed dimensionless.

            >>> print(numerix.round_(float(PhysicalField("2. deg")), 6))
            0.034907

        If the quantity is not dimensionless, the conversion fails.

            >>> float(PhysicalField("2. m"))
            Traceback (most recent call last):
                ...
            TypeError: Not possible to convert a PhysicalField with dimensions to float

        Just as a Numeric_ `array` cannot be cast to float, neither can `PhysicalField` arrays

            >>> float(PhysicalField(((2., 3.), (4., 5.)), "m/m")) # doctest: +ELLIPSIS
            Traceback (most recent call last):
                ...
            TypeError: only ...-1 arrays can be converted to Python scalars

        .. _Numeric: http://www.numpy.org
        """
        if self.unit.isDimensionlessOrAngle():
            return float(self.numericValue)
        else:
            raise TypeError('Not possible to convert a PhysicalField with dimensions to float')

    def __gt__(self, other):
        """
        Compare `self` to `other`, returning an array of Boolean values
        corresponding to the test against each element.

            >>> a = PhysicalField(((3., 4.), (5., 6.)), "m")
            >>> print(numerix.allclose(a > PhysicalField("13 ft"),
            ...                        [[False, True], [ True, True]]))
            True

        Appropriately formatted dimensional quantity strings can also be
        compared.

            >>> print(numerix.allclose(a > "13 ft",
            ...                        [[False, True], [ True, True]]))
            True

        Arrays are compared element to element

            >>> print(numerix.allclose(a > PhysicalField(((3., 13.), (17., 6.)), "ft"),
            ...                        [[ True, True], [False, True]]))
            True

        Units must be compatible

            >>> print(a > PhysicalField("1 lb"))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units

        And so must array dimensions

            >>> print(a > PhysicalField(((3., 13., 4.), (17., 6., 2.)), "ft")) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            ValueError: shape mismatch: objects cannot be broadcast to a single shape
        """
        other = self._inMyUnits(other)
        return self.value > other.value

    def __lt__(self, other):
        other = self._inMyUnits(other)
        return self.value < other.value

    def __le__(self, other):
        other = self._inMyUnits(other)
        return self.value <= other.value

    def __eq__(self, other):
        other = self._inMyUnits(other)
        return self.value == other.value

    __hash__ = object.__hash__

    def __ne__(self, other):
        other = self._inMyUnits(other)
        return self.value != other.value

    def __ge__(self, other):
        other = self._inMyUnits(other)
        return self.value >= other.value

    def __len__(self):
        if type(self.value) in [type(1), type(1.)]:
            return 1
        else:
            return len(self.value)

    def convertToUnit(self, unit):
        """
        Changes the unit to `unit` and adjusts the value such that the
        combination is equivalent.  The new unit is by a string containing
        its name.  The new unit must be compatible with the previous unit
        of the object.

            >>> e = PhysicalField('2.7 Hartree*Nav')
            >>> e.convertToUnit('kcal/mol')
            >>> print(e)
            1694.27557621 kcal/mol
        """
        unit = _findUnit(unit)
        self.value = _convertValue (self.value, self.unit, unit)
        self._unit = unit

    def inRadians(self):
        """
        Converts an angular quantity to radians and returns the numerical value.

            >>> print(PhysicalField(((2., 3.), (4., 5.)), "rad").inRadians())
            [[ 2.  3.]
             [ 4.  5.]]
            >>> print(PhysicalField(((2., 3.), (4., 5.)), "deg").inRadians())
            [[ 0.03490659  0.05235988]
             [ 0.06981317  0.08726646]]

         As a special case, assumes a dimensionless quantity is already in
         radians.

             >>> print(PhysicalField(((2., 3.), (4., 5.))).inRadians())
             [[ 2.  3.]
              [ 4.  5.]]

         It's an error to convert a quantity with non-angular units

            >>> print(PhysicalField(((2., 3.), (4., 5.)), "m").inRadians())
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units

        """
        if self.unit.isDimensionless():
            return self.value
        else:
            return self.inUnitsOf("rad").value

    def inDimensionless(self):
        """
        Returns the numerical value of a dimensionless quantity.

             >>> print(PhysicalField(((2., 3.), (4., 5.))).inDimensionless())
             [[ 2.  3.]
              [ 4.  5.]]

         It's an error to convert a quantity with units

            >>> print(PhysicalField(((2., 3.), (4., 5.)), "m").inDimensionless())
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units

        """
        if self.unit.isDimensionless():
            return self.value
        else:
            raise TypeError('Incompatible units')

    def inUnitsOf(self, *units):
        """
        Returns one or more `PhysicalField` objects that express the same
        physical quantity in different units.  The units are specified by
        strings containing their names.  The units must be compatible with
        the unit of the object.  If one unit is specified, the return value
        is a single `PhysicalField`.

        >>> freeze = PhysicalField('0 degC')
        >>> print(freeze.inUnitsOf('degF').allclose("32.0 degF"))
        1

        If several units are specified, the return value is a tuple of
        `PhysicalField` instances with with one element per unit such that
        the sum of all quantities in the tuple equals the the original
        quantity and all the values except for the last one are integers.
        This is used to convert to irregular unit systems like
        hour/minute/second.  The original object will not be changed.

        >>> t = PhysicalField(314159., 's')
        >>> from builtins import zip
        >>> print(numerix.allclose([e.allclose(v) for (e, v) in zip(t.inUnitsOf('d', 'h', 'min', 's'),
        ...                                                         ['3.0 d', '15.0 h', '15.0 min', '59.0 s'])],
        ...                        True))
        1
        """
        units = list(map(_findUnit, units))
        if len(units) == 1:
            unit = units[0]
            value = _convertValue (self.value, self.unit, unit)
            return self.__class__(value = value, unit = unit)
        else:
            units.sort()
            result = []
            value = self.value
            unit = self.unit
            for i in range(len(units)-1, -1, -1):
                value = value*unit.conversionFactorTo(units[i])
                if i == 0:
                    rounded = value
                else:
                    rounded = _round(value)
                result.append(self.__class__(value = rounded, unit = units[i]))
                value = value - rounded
                unit = units[i]
            return tuple(result)

    def getsctype(self, default=None):
        """

        Returns the NumPy sctype of the underlying array.

            >>> PhysicalField(1, 'm').getsctype() == numerix.NUMERIX.obj2sctype(numerix.array(1))
            True
            >>> PhysicalField(1., 'm').getsctype() == numerix.NUMERIX.obj2sctype(numerix.array(1.))
            True
            >>> PhysicalField((1, 1.), 'm').getsctype() == numerix.NUMERIX.obj2sctype(numerix.array((1., 1.)))
            True

        """

        if not hasattr(self, 'typecode'):
            self.typecode = numerix.obj2sctype(rep=numerix.array(self.numericValue),
                                               default=default)

        return self.typecode

    def _getUnit(self):
        """
        Return the unit object of `self`.

            >>> PhysicalField("1 m").unit
            <PhysicalUnit m>
        """
        return self._unit

    def _setUnit(self, unit):
        """
        Change the unit object of `self` to `unit`

            >>> a = PhysicalField(value="1 m")
            >>> a.unit = ("m**2/s")
            >>> print(a)
            1.0 m**2/s
        """
        self._unit = _findUnit(unit)

    unit = property(_getUnit, _setUnit)

    @property
    def numericValue(self):
        """
        Return the `PhysicalField` without units, after conversion to base SI units.

            >>> print(numerix.round_(PhysicalField("1 inch").numericValue, 6))
            0.0254
        """
        return self.inSIUnits().value

    # Contributed by Berthold Hoellmann
    def inBaseUnits(self):
        """
        Return the quantity with all units reduced to their base SI elements.

            >>> e = PhysicalField('2.7 Hartree*Nav')
            >>> print(e.inBaseUnits().allclose("7088849.01085 kg*m**2/s**2/mol"))
            1
        """
        if self.unit.factor != 1:
            new_value = self.value * self.unit.factor
        else:
            new_value = self.value
        num = ''
        denom = ''
        for i in range(9):
            unit = _base_names[i]
            power = self.unit.powers[i]
            if power < 0:
                denom = denom + '/' + unit
                if power < -1:
                    denom = denom + '**' + str(-power)
            elif power > 0:
                num = num + '*' + unit
                if power > 1:
                    num = num + '**' + str(power)
        if len(num) == 0:
            num = '1'
        else:
            num = num[1:]
        return self.__class__(value = new_value, unit = num + denom)

    def inSIUnits(self):
        """
        Return the quantity with all units reduced to SI-compatible elements.

            >>> e = PhysicalField('2.7 Hartree*Nav')
            >>> print(e.inSIUnits().allclose("7088849.01085 kg*m**2/s**2/mol"))
            1
        """
        if self.unit.factor != 1:
            return self.inBaseUnits()
        else:
            return self

    def isCompatible (self, unit):
        unit = _findUnit (unit)
        return self.unit.isCompatible (unit)

    def arccos(self):
        """
        Return the inverse cosine of the `PhysicalField` in radians

            >>> print(PhysicalField(0).arccos().allclose("1.57079632679 rad"))
            1

        The input `PhysicalField` must be dimensionless

            >>> print(numerix.round_(PhysicalField("1 m").arccos(), 6))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return PhysicalField(value=umath.arccos(self.inDimensionless()), unit = "rad")

    def arccosh(self):
        """
        Return the inverse hyperbolic cosine of the `PhysicalField`

            >>> print(numerix.allclose(PhysicalField(2).arccosh(),
            ...                        1.31695789692))
            1

        The input `PhysicalField` must be dimensionless

            >>> print(numerix.round_(PhysicalField("1. m").arccosh(), 6))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return umath.arccosh(self.inDimensionless())

    def arcsin(self):
        """
        Return the inverse sine of the `PhysicalField` in radians

            >>> print(PhysicalField(1).arcsin().allclose("1.57079632679 rad"))
            1

        The input `PhysicalField` must be dimensionless

            >>> print(numerix.round_(PhysicalField("1 m").arcsin(), 6))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return PhysicalField(value = umath.arcsin(self.inDimensionless()), unit = "rad")

    def sqrt(self):
        """
        Return the square root of the `PhysicalField`

            >>> print(PhysicalField("100. m**2").sqrt())
            10.0 m

        The resulting unit must be integral

            >>> print(PhysicalField("100. m").sqrt())
            Traceback (most recent call last):
                ...
            TypeError: Illegal exponent
        """
        return pow(self, 0.5)

    def sin(self):
        """
        Return the sine of the `PhysicalField`

            >>> print(PhysicalField(numerix.pi/6, "rad").sin())
            0.5
            >>> print(PhysicalField(30., "deg").sin())
            0.5

        The units of the `PhysicalField` must be an angle

            >>> PhysicalField(30., "m").sin()
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return umath.sin(self.inRadians())

    def sinh(self):
        """
        Return the hyperbolic sine of the `PhysicalField`

            >>> PhysicalField(0.).sinh()
            0.0

        The units of the `PhysicalField` must be dimensionless

            >>> PhysicalField(60., "m").sinh()
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return umath.sinh(self.inDimensionless())

    def cos(self):
        """
        Return the cosine of the `PhysicalField`

            >>> print(numerix.round_(PhysicalField(2*numerix.pi/6, "rad").cos(), 6))
            0.5
            >>> print(numerix.round_(PhysicalField(60., "deg").cos(), 6))
            0.5

        The units of the `PhysicalField` must be an angle

            >>> PhysicalField(60., "m").cos()
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return umath.cos(self.inRadians())

    def cosh(self):
        """
        Return the hyperbolic cosine of the `PhysicalField`

            >>> PhysicalField(0.).cosh()
            1.0

        The units of the `PhysicalField` must be dimensionless

            >>> PhysicalField(60., "m").cosh()
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return umath.cosh(self.inDimensionless())

    def tan(self):
        """
        Return the tangent of the `PhysicalField`

            >>> numerix.round_(PhysicalField(numerix.pi/4, "rad").tan(), 6)
            1.0
            >>> numerix.round_(PhysicalField(45, "deg").tan(), 6)
            1.0

        The units of the `PhysicalField` must be an angle

            >>> PhysicalField(45., "m").tan()
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return umath.tan(self.inRadians())

    def tanh(self):
        """
        Return the hyperbolic tangent of the `PhysicalField`

            >>> print(numerix.allclose(PhysicalField(1.).tanh(), 0.761594155956))
            True

        The units of the `PhysicalField` must be dimensionless

            >>> PhysicalField(60., "m").tanh()
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return umath.tanh(self.inDimensionless())

    def arctan2(self, other):
        """
        Return the arctangent of `self` divided by `other` in radians

            >>> print(numerix.round_(PhysicalField(2.).arctan2(PhysicalField(5.)), 6))
            0.380506

        The input `PhysicalField` objects must be in the same dimensions

            >>> print(numerix.round_(PhysicalField(2.54, "cm").arctan2(PhysicalField(1., "inch")), 6))
            0.785398

            >>> print(numerix.round_(PhysicalField(2.).arctan2(PhysicalField("5. m")), 6))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return PhysicalField(value=umath.arctan2(self.value,
                                                 self._inMyUnits(other).value), unit="rad")

    def arctan(self):
        """
        Return the arctangent of the `PhysicalField` in radians

            >>> print(numerix.round_(PhysicalField(1).arctan(), 6))
            0.785398

        The input `PhysicalField` must be dimensionless

            >>> print(numerix.round_(PhysicalField("1 m").arctan(), 6))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return PhysicalField(value= umath.arctan(self.inDimensionless()), unit = "rad")

    def arctanh(self):
        """
        Return the inverse hyperbolic tangent of the `PhysicalField`

            >>> print(PhysicalField(0.5).arctanh())
            0.549306144334

        The input `PhysicalField` must be dimensionless

            >>> print(numerix.round_(PhysicalField("1 m").arctanh(), 6))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return umath.arctanh(self.inDimensionless())


    def log(self):
        """
        Return the natural logarithm of the `PhysicalField`

            >>> print(numerix.round_(PhysicalField(10).log(), 6))
            2.302585

        The input `PhysicalField` must be dimensionless

            >>> print(numerix.round_(PhysicalField("1. m").log(), 6))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return umath.log(self.inDimensionless())

    def log10(self):
        """
        Return the base-10 logarithm of the `PhysicalField`

            >>> print(numerix.round_(PhysicalField(10.).log10(), 6))
            1.0

        The input `PhysicalField` must be dimensionless

            >>> print(numerix.round_(PhysicalField("1. m").log10(), 6))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        return umath.log10(self.inDimensionless())


    def floor(self):
        """
        Return the largest integer less than or equal to the `PhysicalField`.

            >>> print(PhysicalField(2.2, "m").floor())
            2.0 m
        """
        return self.__class__(value = umath.floor(self.value), unit = self.unit)

    def ceil(self):
        """
        Return the smallest integer greater than or equal to the `PhysicalField`.

            >>> print(PhysicalField(2.2, "m").ceil())
            3.0 m
        """
        return self.__class__(value = umath.ceil(self.value), unit = self.unit)

    def conjugate(self):
        """
        Return the complex conjugate of the `PhysicalField`.

            >>> print(PhysicalField(2.2 - 3j, "ohm").conjugate() == PhysicalField(2.2 + 3j, "ohm"))
            True
        """
        return self.__class__(value = umath.conjugate(self.value), unit = self.unit)

    def dot(self, other):
        """
        Return the dot product of `self` with `other`. The resulting
        unit is the product of the units of `self` and `other`.

            >>> v = PhysicalField(((5., 6.), (7., 8.)), "m")
            >>> print(PhysicalField(((1., 2.), (3., 4.)), "m").dot(v))
            [ 26.  44.] m**2

        """
        if not isinstance(other, PhysicalField):
            return self.__class__(value = numerix.dot(self.value, other), unit = self.unit)
        value = numerix.dot(self.value, other.value)
        unit = self.unit*other.unit
        if unit.isDimensionless():
            return value*unit.factor
        else:
            return self.__class__(value = value, unit = unit)

    def take(self, indices, axis = 0):
        """
        Return the elements of `self` specified by the elements of `indices`.
        The resulting `PhysicalField` array has the same units as the original.

            >>> print(PhysicalField((1., 2., 3.), "m").take((2, 0)))
            [ 3.  1.] m

        The optional third argument specifies the axis along which the selection
        occurs, and the default value (as in the example above) is 0, the first axis.

            >>> print(PhysicalField(((1., 2., 3.), (4., 5., 6.)), "m").take((2, 0), axis = 1))
            [[ 3.  1.]
             [ 6.  4.]] m

        """
        return self.__class__(value = numerix.take(self.value, indices, axis), unit = self.unit)

    def put(self, indices, values):
        """
        `put` is the opposite of `take`. The values of `self` at the locations
        specified in `indices` are set to the corresponding value of `values`.

        The `indices` can be any integer sequence object with values suitable for
        indexing into the flat form of `self`. The `values` must be any sequence of
        values that can be converted to the typecode of `self`.

            >>> f = PhysicalField((1., 2., 3.), "m")
            >>> f.put((2, 0), PhysicalField((2., 3.), "inch"))
            >>> print(f)
            [ 0.0762  2.      0.0508] m

        The units of `values` must be compatible with `self`.

            >>> f.put(1, PhysicalField(3, "kg"))
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units
        """
        numerix.put(self.value, indices, self._inMyUnits(values).value)

    @property
    def shape(self):
        """Tuple of array dimensions."""
        from fipy.tools import numerix
        return numerix.getShape(self.value)

    @property
    def itemsize(self):
        return self.value.itemsize

    def reshape(self, shape):
        """
        Changes the shape of `self` to that specified in `shape`

            >>> print(PhysicalField((1., 2., 3., 4.), "m").reshape((2, 2)))
            [[ 1.  2.]
             [ 3.  4.]] m

        The new shape must have the same size as the existing one.

            >>> print(PhysicalField((1., 2., 3., 4.), "m").reshape((2, 3))) # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            ValueError: total size of new array must be unchanged
        """
        return self.__class__(value = numerix.reshape(self.value, shape), unit = self.unit)

    def sum(self, index = 0):
        """
        Returns the sum of all of the elements in `self` along the
        specified axis (first axis by default).

            >>> print(PhysicalField(((1., 2.), (3., 4.)), "m").sum())
            [ 4.  6.] m
            >>> print(PhysicalField(((1., 2.), (3., 4.)), "m").sum(1))
            [ 3.  7.] m
        """
        return self.__class__(value = numerix.sum(self.value, index), unit = self.unit)

    def allclose(self, other, atol = None, rtol = 1.e-8):
        """
        This function tests whether or not `self` and `other` are equal subject
        to the given relative and absolute tolerances.  The formula used is::

            | self - other | < atol + rtol * | other |

        This means essentially that both elements are small compared to `atol` or
        their difference divided by `other`'s value is small compared to `rtol`.
        """
        other = self._inMyUnits(other)
        if atol is None:
            atol = PhysicalField(1.e-5, self.unit)
        else:
            atol = self._inMyUnits(atol)

        return MA.allclose(self.value, other.value, atol = atol.value, rtol = rtol)

    def allequal(self, other):
        """
        This function tests whether or not `self` and `other` are exactly equal.
        """
        other = self._inMyUnits(other)
        return MA.allequal(self.value, other.value)

class PhysicalUnit(object):
    """
    A `PhysicalUnit` represents the units of a `PhysicalField`.
    """

    def __init__(self, names, factor, powers, offset=0):
        """
        This class is not generally not instantiated by users of this module,
        but rather it is created in the process of constructing a `PhysicalField`.

        Parameters
        ----------
        names : str
            Name of the unit
        factor : float
            Multiplier between the unit and the fundamental SI unit
        powers`: `array_like` of :obj:`float`
            Nine elements representing the fundamental SI units of
            ["m", "kg", "s", "A", "K", "mol", "cd", "rad", "sr"]
        offset : float
            Displacement between the zero-point of the unit and the
            zero-point of the corresponding fundamental SI unit.
        """
        if isinstance(names, string_types):
            self.names = _NumberDict()
            self.names[names] = 1
        else:
            self.names = names
        self.factor = factor
        self.offset = offset
        self.powers = numerix.array(powers)

    def __repr__(self):
        """
        Return representation of a physical unit

            >>> PhysicalUnit('m',   1.,    [1, 0, 0, 0, 0, 0, 0, 0, 0])
            <PhysicalUnit m>
        """
        return '<PhysicalUnit ' + self.name() + '>'

    __str__ = __repr__

    def _checkSame(self, other):
        if not isinstance(other, PhysicalUnit):
            if other == 1:
                return self.isDimensionless()
            else:
                raise TypeError('PhysicalUnits can only be compared with other PhysicalUnits')
        if not numerix.alltrue(self.powers == other.powers):
            raise TypeError('Incompatible units')

    def __eq__(self, other):
        """
        Determine if units are identical

        >>> a = PhysicalField("1. m")
        >>> b = PhysicalField("3. ft")
        >>> a.unit == b.unit
        0
        >>> a.unit == b.inBaseUnits().unit
        1

        Units can only be compared with other units

        >>> a.unit == 3
        Traceback (most recent call last):
            ...
        TypeError: PhysicalUnits can only be compared with other PhysicalUnits
        """
        self._checkSame(other)
        return self.factor == other.factor

    def __lt__(self, other):
        self._checkSame(other)
        return self.factor < other.factor

    def __le__(self, other):
        self._checkSame(other)
        return self.factor <= other.factor

    def __ne__(self, other):
        self._checkSame(other)
        return self.factor != other.factor

    def __gt__(self, other):
        self._checkSame(other)
        return self.factor > other.factor

    def __ge__(self, other):
        self._checkSame(other)
        return self.factor >= other.factor

    def __mul__(self, other):
        """
        Multiply units together

            >>> a = PhysicalField("1. m")
            >>> b = PhysicalField("3. ft")
            >>> a.unit * b.unit == _findUnit('ft*m')
            True
            >>> a.unit * b.inBaseUnits().unit
            <PhysicalUnit m**2>
            >>> c = PhysicalField("1. s")
            >>> d = PhysicalField("3. Hz")
            >>> c.unit * d.unit == _findUnit('Hz*s')
            True
            >>> c.unit * d.inBaseUnits().unit
            <PhysicalUnit 1>

        or multiply units by numbers

            >>> a.unit * 3.
            <PhysicalUnit m*3.0>

        Units must have zero offset to be multiplied

            >>> e = PhysicalField("1. kB")
            >>> f = PhysicalField("25. degC")
            >>> e.unit * f.unit
            Traceback (most recent call last):
                ...
            TypeError: cannot multiply units with non-zero offset
            >>> e.unit * f.inBaseUnits().unit
            <PhysicalUnit kB*K>
        """
        if self.offset != 0 or (isinstance(other, PhysicalUnit) and other.offset != 0):
            raise TypeError("cannot multiply units with non-zero offset")
        if isinstance(other, PhysicalUnit):
            return PhysicalUnit(self.names+other.names,
                                self.factor*other.factor,
                                self.powers + other.powers)
        else:
            return PhysicalUnit(self.names+{str(other): 1},
                                self.factor*other,
                                self.powers,
                                self.offset * other)

    __rmul__ = __mul__

    def __truediv__(self, other):
        """
        Divide one unit by another

            >>> a = PhysicalField("1. m")
            >>> b = PhysicalField("3. ft")
            >>> a.unit / b.unit
            <PhysicalUnit m/ft>
            >>> a.unit / b.inBaseUnits().unit
            <PhysicalUnit 1>
            >>> c = PhysicalField("1. s")
            >>> d = PhysicalField("3. Hz")
            >>> c.unit / d.unit
            <PhysicalUnit s/Hz>
            >>> c.unit / d.inBaseUnits().unit
            <PhysicalUnit s**2/1>

        or divide units by numbers

            >>> a.unit / 3.
            <PhysicalUnit m/3.0>

        Units must have zero offset to be divided

            >>> e = PhysicalField("1. J")
            >>> f = PhysicalField("25. degC")
            >>> e.unit / f.unit
            Traceback (most recent call last):
                ...
            TypeError: cannot divide units with non-zero offset
            >>> e.unit / f.inBaseUnits().unit
            <PhysicalUnit J/K>
        """
        if self.offset != 0 or (isinstance(other, PhysicalUnit) and other.offset != 0):
            raise TypeError("cannot divide units with non-zero offset")
        if isinstance(other, PhysicalUnit):
            return PhysicalUnit(self.names-other.names,
                                self.factor / other.factor,
                                self.powers - other.powers)
        else:
            return PhysicalUnit(self.names+{str(other): -1},
                                self.factor / other, self.powers)

    __div__ = __truediv__

    def __rtruediv__(self, other):
        """
        Divide something by a unit

            >>> a = PhysicalField("1. m")
            >>> 3. / a.unit
            <PhysicalUnit 3.0/m>

        Units must have zero offset to be divided

            >>> b = PhysicalField("25. degC")
            >>> 3. / b.unit
            Traceback (most recent call last):
                ...
            TypeError: cannot divide units with non-zero offset
            >>> 3. / b.inBaseUnits().unit
            <PhysicalUnit 3.0/K>
        """
        if self.offset != 0 or (isinstance(other, PhysicalUnit) and other.offset != 0):
            raise TypeError("cannot divide units with non-zero offset")
        if isinstance(other, PhysicalUnit):
            return PhysicalUnit(other.names-self.names,
                                other.factor / self.factor,
                                list(map(lambda a, b: a-b,
                                    other.powers, self.powers)))
        else:
            return PhysicalUnit({str(other): 1}-self.names,
                                other / self.factor,
                                -self.powers)

    __rdiv__ = __rtruediv__

    def __pow__(self, other):
        """
        Raise a unit to an integer power

            >>> a = PhysicalField("1. m")
            >>> a.unit**2
            <PhysicalUnit m**2>
            >>> a.unit**-2
            <PhysicalUnit 1/m**2>

        Non-integer powers are not supported

            >>> a.unit**0.5
            Traceback (most recent call last):
                ...
            TypeError: Illegal exponent

        Units must have zero offset to be exponentiated

            >>> b = PhysicalField("25. degC")
            >>> b.unit**2
            Traceback (most recent call last):
                ...
            TypeError: cannot exponentiate units with non-zero offset
            >>> b.inBaseUnits().unit**2
            <PhysicalUnit K**2>
        """
        if self.offset != 0:
            raise TypeError("cannot exponentiate units with non-zero offset")
        if isinstance(other, type(0)):
            return PhysicalUnit(other*self.names, pow(self.factor, other),
                                self.powers*other)

        other = float(other)

        rounded = int(umath.floor(other+0.5))
        if abs(other-rounded) < 1.e-10:
            other = int(other)
            return PhysicalUnit(other*self.names, pow(self.factor, other),
                                self.powers*other)

        inv_exp = 1./other
        rounded = int(umath.floor(inv_exp+0.5))
        if abs(inv_exp-rounded) < 1.e-10:
            if numerix.logical_and.reduce(self.powers % rounded == 0):
                f = pow(self.factor, other)
                p = self.powers // rounded
                if reduce(lambda a, b: a and b,
                          list(map(lambda x, e=rounded: x%e == 0,
                              list(self.names.values())))):
                    names = self.names // rounded
                else:
                    names = _NumberDict()
                    if f != 1.:
                        names[str(f)] = 1
                    for i in range(len(p)):
                        names[_base_names[i]] = p[i]
                return PhysicalUnit(names, f, p)
            else:
                raise TypeError('Illegal exponent')
        raise TypeError('Only integer and inverse integer exponents allowed')

    def conversionFactorTo(self, other):
        """
        Return the multiplication factor between two physical units

            >>> a = PhysicalField("1. mm")
            >>> b = PhysicalField("1. inch")
            >>> print(numerix.round_(b.unit.conversionFactorTo(a.unit), 6))
            25.4

        Units must have the same fundamental SI units

            >>> c = PhysicalField("1. K")
            >>> c.unit.conversionFactorTo(a.unit)
            Traceback (most recent call last):
                ...
            TypeError: Incompatible units

        If units have different offsets, they must have the same factor

            >>> d = PhysicalField("1. degC")
            >>> c.unit.conversionFactorTo(d.unit)
            1.0
            >>> e = PhysicalField("1. degF")
            >>> c.unit.conversionFactorTo(e.unit)
            Traceback (most recent call last):
                ...
            TypeError: Unit conversion (K to degF) cannot be expressed as a simple multiplicative factor
        """
        if not numerix.alltrue(self.powers == other.powers):
            if self.isDimensionlessOrAngle() and other.isDimensionlessOrAngle():
                return self.factor / other.factor
            else:
                raise TypeError('Incompatible units')
        if self.offset != other.offset and self.factor != other.factor:
            raise TypeError(('Unit conversion (%s to %s) cannot be expressed ' +
                   'as a simple multiplicative factor') % \
                  (self.name(), other.name()))
        return self.factor / other.factor

    def conversionTupleTo(self, other): # added 1998/09/29 GPW
        """
        Return a `tuple` of the multiplication factor and offset between two physical units

            >>> a = PhysicalField("1. K").unit
            >>> b = PhysicalField("1. degF").unit
            >>> from builtins import str
            >>> [str(numerix.round_(element, 6)) for element in b.conversionTupleTo(a)]
            ['0.555556', '459.67']
        """
        if not numerix.alltrue(self.powers == other.powers):
            raise TypeError('Incompatible units')

        # let (s1,d1) be the conversion tuple from 'self' to base units
        #   (ie. (x+d1)*s1 converts a value x from 'self' to base units,
        #   and (x/s1)-d1 converts x from base to 'self' units)
        # and (s2,d2) be the conversion tuple from 'other' to base units
        # then we want to compute the conversion tuple (S,D) from
        #   'self' to 'other' such that (x+D)*S converts x from 'self'
        #   units to 'other' units
        # the formula to convert x from 'self' to 'other' units via the
        #   base units is (by definition of the conversion tuples):
        #     ( ((x+d1)*s1) / s2 ) - d2
        #   = ( (x+d1) * s1/s2) - d2
        #   = ( (x+d1) * s1/s2 ) - (d2*s2/s1) * s1/s2
        #   = ( (x+d1) - (d1*s2/s1) ) * s1/s2
        #   = (x + d1 - d2*s2/s1) * s1/s2
        # thus, D = d1 - d2*s2/s1 and S = s1/s2
        factor = self.factor / other.factor
        offset = self.offset - (other.offset * other.factor / self.factor)
        return (factor, offset)

    def isCompatible (self, other):     # added 1998/10/01 GPW
        """
        Returns a list of which fundamental SI units are compatible between
        `self` and `other`

            >>> a = PhysicalField("1. mm")
            >>> b = PhysicalField("1. inch")
            >>> print(numerix.allclose(a.unit.isCompatible(b.unit),
            ...                        [True, True, True, True, True, True, True, True, True]))
            True
            >>> c = PhysicalField("1. K")
            >>> print(numerix.allclose(a.unit.isCompatible(c.unit),
            ...                        [False, True, True, True, False, True, True, True, True]))
            True

        """
        return self.powers == other.powers

    def isDimensionless(self):
        """
        Returns `True` if the unit is dimensionless

            >>> PhysicalField("1. m/m").unit.isDimensionless()
            1
            >>> PhysicalField("1. inch").unit.isDimensionless()
            0
        """
        return not numerix.logical_or.reduce(self.powers)

    def isAngle(self):
        """
        Returns `True` if the unit is an angle

            >>> PhysicalField("1. deg").unit.isAngle()
            1
            >>> PhysicalField("1. rad").unit.isAngle()
            1
            >>> PhysicalField("1. inch").unit.isAngle()
            0
        """
        return self.powers[7] == 1 and \
               numerix.add.reduce(self.powers) == 1

    def isInverseAngle(self):
        """
        Returns `True` if the 1 divided by the unit is an angle

            >>> PhysicalField("1. deg**-1").unit.isInverseAngle()
            1
            >>> PhysicalField("1. 1/rad").unit.isInverseAngle()
            1
            >>> PhysicalField("1. inch").unit.isInverseAngle()
            0
        """
        return self.powers[7] == -1 and \
               numerix.add.reduce(self.powers) == -1


    def isDimensionlessOrAngle(self):
        """
        Returns `True` if the unit is dimensionless or an angle

            >>> PhysicalField("1. m/m").unit.isDimensionlessOrAngle()
            1
            >>> PhysicalField("1. deg").unit.isDimensionlessOrAngle()
            1
            >>> PhysicalField("1. rad").unit.isDimensionlessOrAngle()
            1
            >>> PhysicalField("1. inch").unit.isDimensionlessOrAngle()
            0
        """
        return self.isDimensionless() or self.isAngle()

    def setName(self, name):
        """
        Set the name of the unit to `name`

            >>> a = PhysicalField("1. m/s").unit
            >>> a
            <PhysicalUnit m/s>
            >>> a.setName('meterpersecond')
            >>> a
            <PhysicalUnit meterpersecond>
        """
        self.names = _NumberDict()
        self.names[name] = 1

    def name(self):
        """
        Return the name of the unit

            >>> PhysicalField("1. m").unit.name()
            'm'
            >>> (PhysicalField("1. m") / PhysicalField("1. s")
            ...  / PhysicalField("1. s")).unit.name()
            'm/s**2'
        """
        num = ''
        denom = ''
        for unit in list(self.names.keys()):
            power = self.names[unit]
            if power < 0:
                denom = denom + '/' + unit
                if power < -1:
                    denom = denom + '**' + str(-power)
            elif power > 0:
                num = num + '*' + unit
                if power > 1:
                    num = num + '**' + str(power)
        if len(num) == 0:
            num = '1'
        else:
            num = num[1:]
        return text_to_native_str(num + denom)

# Helper functions

def _findUnit(unit):
    """
    Return the `PhysicalUnit` corresponding to `unit`

        >>> _findUnit('m')
        <PhysicalUnit m>
        >>> _findUnit('')
        <PhysicalUnit 1>
        >>> _findUnit(1.)
        <PhysicalUnit 1>
        >>> _findUnit(PhysicalField("4 N*m").unit) == _findUnit('m*N')
        True
        >>> _findUnit(2.)
        Traceback (most recent call last):
            ...
        TypeError: 2.0 is not a unit
    """
##     print unit, type(unit)

    if isinstance(unit, string_types):
        name = unit.strip()
        if len(name) == 0 or unit == '1':
            unit = _unity
        else:
            unit = eval(name, _unit_table)
            for cruft in ['__builtins__', '__args__']:
                try: del _unit_table[cruft]
                except: pass

    if not isinstance(unit, PhysicalUnit):
        if unit == 1:
            unit = _unity
        else:
            raise TypeError(str(unit) + ' is not a unit')
    return unit

def _round(x):
    if umath.greater(x, 0.):
        return umath.floor(x)
    else:
        return umath.ceil(x)


def _convertValue (value, src_unit, target_unit):
    (factor, offset) = src_unit.conversionTupleTo(target_unit)
    return (value + offset) * factor

def _Scale(quantity, scaling):
    """
    Normalize `quantity` by `scaling`.

    `quantity` can be a `PhysicalField`

        >>> print(numerix.round_(_Scale(PhysicalField("1. inch"), PhysicalField("1. mm")), 6))
        25.4

    or a value-unit string convertible to a `PhysicalField`

        >>> print(numerix.round_(_Scale("1. inch", PhysicalField("1. mm")), 6))
        25.4

    or a dimensionless number. A dimensionless number is left alone.

        >>> print(numerix.round_(_Scale(PhysicalField(2.), PhysicalField("1. mm")), 6))
        2.0

    It is an error for the result to have dimensions.

        >>> print(_Scale(PhysicalField("1. s"), PhysicalField("1. mm")))
        Traceback (most recent call last):
            ...
        TypeError: <PhysicalUnit s> and <PhysicalUnit m> are incompatible
    """

    quantity = PhysicalField(quantity)

    if not quantity.unit.isDimensionless():
        scaling = PhysicalField(scaling)
        # normalize quantity to scaling
        # error will be thrown if incompatible
        dimensionless = quantity / scaling
    else:
        # Assume quantity is a dimensionless number and return it.
        # Automatically throws an error if it's not a number.
        dimensionless = quantity

    if isinstance(dimensionless, PhysicalField) and not dimensionless.unit.isDimensionless():
        raise TypeError(repr(quantity.inBaseUnits().unit) + ' and ' \
        + repr(scaling.inBaseUnits().unit) \
        + ' are incompatible')

    return dimensionless

def _isVariable(var):
    from fipy.variables import variable
    return isinstance(var, variable.Variable)

# SI unit definitions

_base_names = ['m', 'kg', 's', 'A', 'K', 'mol', 'cd', 'rad', 'sr']

_base_units = [('m',   PhysicalUnit('m',   1.,    [1, 0, 0, 0, 0, 0, 0, 0, 0])),
               ('g',   PhysicalUnit('g',   0.001, [0, 1, 0, 0, 0, 0, 0, 0, 0])),
               ('s',   PhysicalUnit('s',   1.,    [0, 0, 1, 0, 0, 0, 0, 0, 0])),
               ('A',   PhysicalUnit('A',   1.,    [0, 0, 0, 1, 0, 0, 0, 0, 0])),
               ('K',   PhysicalUnit('K',   1.,    [0, 0, 0, 0, 1, 0, 0, 0, 0])),
               ('mol', PhysicalUnit('mol', 1.,    [0, 0, 0, 0, 0, 1, 0, 0, 0])),
               ('cd',  PhysicalUnit('cd',  1.,    [0, 0, 0, 0, 0, 0, 1, 0, 0])),
               ('rad', PhysicalUnit('rad', 1.,    [0, 0, 0, 0, 0, 0, 0, 1, 0])),
               ('sr',  PhysicalUnit('sr',  1.,    [0, 0, 0, 0, 0, 0, 0, 0, 1])),
               ]

_prefixes = [('Y',  1.e24),
             ('Z',  1.e21),
             ('E',  1.e18),
             ('P',  1.e15),
             ('T',  1.e12),
             ('G',  1.e9),
             ('M',  1.e6),
             ('k',  1.e3),
             ('h',  1.e2),
             ('da', 1.e1),
             ('d',  1.e-1),
             ('c',  1.e-2),
             ('m',  1.e-3),
             ('mu', 1.e-6),
             ('n',  1.e-9),
             ('p',  1.e-12),
             ('f',  1.e-15),
             ('a',  1.e-18),
             ('z',  1.e-21),
             ('y',  1.e-24),
             ]

_unit_table = {}

for unit in _base_units:
    _unit_table[unit[0]] = unit[1]

def _addUnit(name, unit):
    if name in _unit_table:
        raise KeyError('Unit ' + name + ' already defined')
    if isinstance(unit, string_types):
        unit = eval(unit, _unit_table)
        for cruft in ['__builtins__', '__args__']:
            try: del _unit_table[cruft]
            except: pass
    unit.setName(name)
    _unit_table[name] = unit

def _addPrefixed(unit):
    for prefix in _prefixes:
        name = prefix[0] + unit
        _addUnit(name, prefix[1]*_unit_table[unit])


# SI derived units; these automatically get prefixes

_unit_table['kg'] = PhysicalUnit('kg',   1., [0, 1, 0, 0, 0, 0, 0, 0, 0])

_addUnit('Hz', '1/s')                # Hertz
_addUnit('N', 'm*kg/s**2')           # Newton
_addUnit('Pa', 'N/m**2')             # Pascal
_addUnit('J', 'N*m')                 # Joule
_addUnit('W', 'J/s')                 # Watt
_addUnit('C', 's*A')                 # Coulomb
_addUnit('V', 'W/A')                 # Volt
_addUnit('F', 'C/V')                 # Farad
_addUnit('ohm', 'V/A')               # Ohm
_addUnit('S', 'A/V')                 # Siemens
_addUnit('Wb', 'V*s')                # Weber
_addUnit('T', 'Wb/m**2')             # Tesla
_addUnit('H', 'Wb/A')                # Henry
_addUnit('lm', 'cd*sr')              # Lumen
_addUnit('lx', 'lm/m**2')            # Lux
_addUnit('Bq', '1/s')                # Becquerel
_addUnit('Gy', 'J/kg')               # Gray
_addUnit('Sv', 'J/kg')               # Sievert

del _unit_table['kg']

for unit in list(_unit_table.keys()):
    _addPrefixed(unit)

# Fundamental constants

_unit_table['pi'] = umath.pi
_addUnit('c', '299792458.*m/s')      # speed of light
_addUnit('mu0', '4.e-7*pi*N/A**2')   # permeability of vacuum
_addUnit('eps0', '1/mu0/c**2')       # permittivity of vacuum
_addUnit('Grav', '6.6742e-11*m**3/kg/s**2') # gravitational constant
_addUnit('hplanck', '6.6260693e-34*J*s')     # Planck constant
_addUnit('hbar', 'hplanck/(2*pi)')   # Planck constant / 2pi
_addUnit('e', '1.60217653e-19*C')    # elementary charge
_addUnit('me', '9.1093826e-31*kg')   # electron mass
_addUnit('mp', '1.67262171e-27*kg')  # proton mass
_addUnit('Nav', '6.0221415e23/mol')  # Avogadro number
_addUnit('kB', '1.3806505e-23*J/K')  # Boltzmann constant

_addUnit('gn', '9.80665*m/s**2')     # standard gravitational acceleration
# Time units

_addUnit('min', '60*s')              # minute
_addUnit('h', '60*min')              # hour
_addUnit('d', '24*h')                # day
_addUnit('wk', '7*d')                # week
_addUnit('yr', '365*d')              # year
_addUnit('yrJul', '365.25*d')        # Julian year
_addUnit('yrSid', '365.2564*d')      # sidereal year

# Length units

_addUnit('inch', '2.54*cm')          # inch
_addUnit('ft', '12*inch')            # foot
_addUnit('yd', '3*ft')               # yard
_addUnit('mi', '5280.*ft')           # (British) mile
_addUnit('nmi', '1852.*m')           # Nautical mile
_addUnit('Ang', '1.e-10*m')          # Angstrom
_addUnit('lyr', 'c*yrJul')           # light year
_addUnit('Bohr', '4*pi*eps0*hbar**2/me/e**2')  # Bohr radius

# Area units

_addUnit('ha', '10000*m**2')         # hectare
_addUnit('acres', 'mi**2/640')       # acre
_addUnit('b', '1.e-28*m')            # barn

# Volume units

_addUnit('l', 'dm**3')               # liter
_addUnit('dl', '0.1*l')
_addUnit('cl', '0.01*l')
_addUnit('ml', '0.001*l')
_addUnit('tsp', '4.928922*ml')  # teaspoon
_addUnit('tbsp', '3*tsp')            # tablespoon
_addUnit('floz', '2*tbsp')           # fluid ounce
_addUnit('cup', '8*floz')            # cup
_addUnit('pt', '16*floz')            # pint
_addUnit('qt', '2*pt')               # quart
_addUnit('galUS', '4*qt')            # US gallon
_addUnit('galUK', '4.54609*l')       # British gallon

# Mass units

_addUnit('amu', '1.6605402e-27*kg')  # atomic mass units
_addUnit('lb', '4.5359237e-1*kg')    # pound
_addUnit('oz', 'lb/16')              # ounce
_addUnit('ton', '2000*lb')           # ton

# Force units

_addUnit('dyn', '1.e-5*N')           # dyne (cgs unit)

# Energy units

_addUnit('erg', '1.e-7*J')           # erg (cgs unit)
_addUnit('eV', 'e*V')                # electron volt
_addPrefixed('eV')
_addUnit('Hartree', 'me*e**4/16/pi**2/eps0**2/hbar**2')
_addUnit('invcm', 'hplanck*c/cm')    # Wavenumbers/inverse cm
_addUnit('Ken', 'kB*K')              # Kelvin as energy unit
_addUnit('cal', '4.184*J')           # thermochemical calorie
_addUnit('kcal', '1000*cal')         # thermochemical kilocalorie
_addUnit('cali', '4.1868*J')         # international calorie
_addUnit('kcali', '1000*cali')       # international kilocalorie
_addUnit('Btui', '1055.05585262*J')  # international British thermal unit

# Power units

_addUnit('hpEl', '746*W')            # electric horsepower
_addUnit('hpUK', '745.7*W')            # horsepower

# Pressure units

_addUnit('bar', '1.e5*Pa')           # bar (cgs unit)
_addUnit('atm', '101325.*Pa')        # standard atmosphere
_addUnit('Torr', 'atm/760')          # torr = mm of mercury
_addUnit('psi', 'lb*gn/inch**2')     # pounds per square inch

# Angle units

_addUnit('deg', 'pi*rad/180')        # degrees

# Temperature units -- can't use the 'eval' trick that _addUnit provides
# for degC and degF because you can't add units
kelvin = _findUnit ('K')
_addUnit ('degR', '(5./9.)*K')       # degrees Rankine
_addUnit ('degC', PhysicalUnit (None, 1.0, kelvin.powers, 273.15))
_addUnit ('degF', PhysicalUnit (None, 5./9., kelvin.powers, 459.67))
del kelvin

_unity = eval("m/m", _unit_table)

def _getUnitStrings():

    working_table = _unit_table.copy()

    def _getSortedUnitStrings(unitDict):
        strings = []
        keys = [str(key) for key in list(unitDict.keys())]
        keys.sort(key=str.lower)
        for key in keys:
            if key in unitDict:
                unit = unitDict[key]
                if isinstance(unit, PhysicalUnit):
                    tmp = PhysicalField(value = 1, unit = unit)
                    strings.append("%10s = %s" % (str(tmp), str(tmp.inBaseUnits())))

                    del working_table[key]

        return strings

    def _deleteFactors(unit):
        for prefix, factor in _prefixes:
            if (prefix + unit.name()) in working_table:
                del working_table[prefix + unit.name()]


    units = []

    units.append("\nBase SI units::\n")
    units.append("\t%s" % ", ".join(_base_names))
    for name in _base_names:
        if name in working_table:
            del working_table[name]
    for name, unit in _base_units:
        _deleteFactors(unit)

    units.append("\nSI prefixes::\n")
    for prefix, factor in _prefixes:
        units.append("%10s = %g" % (prefix, factor))

    units.append("\nUnits derived from SI (accepting SI prefixes)::\n")
    derived = {}
    for key in list(working_table.keys()):
        if key in working_table:
            unit = working_table[key]
            if isinstance(unit, PhysicalUnit) and unit.factor == 1:
                derived[unit.name()] = unit
                _deleteFactors(unit)

    units.extend(_getSortedUnitStrings(derived))

    units.append("\nOther units that accept SI prefixes::\n")
    prefixed = {}
    for key in list(working_table.keys()):
        if key in working_table:
            unit = working_table[key]
            isPrefixed = 1
            if isinstance(unit, PhysicalUnit):
                for prefix, factor in _prefixes:
                    if (prefix + key) not in working_table:
                        isPrefixed = 0
                        break
                if isPrefixed:
                    prefixed[unit.name()] = unit
                    _deleteFactors(unit)

    units.extend(_getSortedUnitStrings(prefixed))

    units.append("\nAdditional units and constants::\n")
    units.extend(_getSortedUnitStrings(working_table))

    return "\n".join(units)

__doc__ += _getUnitStrings()

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
##     print _getUnitStrings()
    _test()



from __future__ import unicode_literals
from builtins import object
__docformat__ = 'restructuredtext'

from fipy.tools import numerix
from fipy.variables.variable import Variable
from fipy.tools.dimensions.physicalField import PhysicalField

__all__ = ["BoundaryCondition"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class BoundaryCondition(object):
    """
    Generic boundary condition base class.

    .. attention:: This class is abstract. Always create one of its subclasses.
    """

    def __init__(self, faces, value):
        """
        Parameters
        ----------
        faces : :obj:`~fipy.variables.faceVariable.FaceVariable` of :obj:`bool`
            Mask of faces where this condition applies.
        value : float
            Value to impose.
        """
        if self.__class__ is BoundaryCondition:
            raise NotImplementedError("can't instantiate abstract base class")

        self.faces = faces
        if not (isinstance(value, PhysicalField) or isinstance(value, Variable)):
            value = PhysicalField(value)
        self.value = value

        if not (self.faces | self.faces.mesh.exteriorFaces
                == self.faces.mesh.exteriorFaces).value.all():
            raise IndexError('Face list has interior faces')

        self.adjacentCellIDs = self.faces.mesh._adjacentCellIDs[0][self.faces.value]
        self.boundaryConditionApplied = False

    def _buildMatrix(self, SparseMatrix, Ncells, MaxFaces, coeff):
        """Return the effect of this boundary condition on the equation
        solution matrices.

        `_buildMatrix()` is called by each `Term` of each `Equation`.

        A `tuple` of (`LL`, `bb`) is calculated, to be added to the Term's
        (**L**, **b**) matrices.

        Parameters
        ----------
        SparseMatrix : ~fipy.matrices.sparseMatrix._SparseMatrix
        Ncells : int
            Size of matrices
        MaxFaces : int
            Maximum number of faces per cell (determines bandwidth of
            :math:`\mathsf{L}`)
        coeff : list
            Contribution due to this face
        """
        raise NotImplementedError

    def _getDerivative(self, order):
        """Return a tuple of the boundary conditions to apply
        to the term and to the derivative of the term
        """
        if order == 0:
            return self
        else:
            return None

    def __repr__(self):
        return "%s(faces = %s, value = %s)" % (self.__class__.__name__, repr(self.faces), repr(self.value))

    def _resetBoundaryConditionApplied(self):
        self.boundaryConditionApplied = False

    def _test(self):
        """
        The `BoundaryCondition` class should raise an error when
        invoked with internal faces. Don't use the `BoundaryCondition`
        class in this manner. This is merely a test.

        >>> from fipy.meshes import Grid1D
        >>> mesh = Grid1D(nx = 2)
        >>> from fipy.tools import parallelComm
        >>> if parallelComm.procID == 0:
        ...     bc = __BoundaryCondition(mesh.interiorFaces, 0)
        ... else:
        ...     raise IndexError("Face list has interior faces")
        Traceback (most recent call last):
            ...
        IndexError: Face list has interior faces
        """
        pass

class __BoundaryCondition(BoundaryCondition):
    """
    Dummy subclass for tests
    """
    pass

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()



from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.variables.gaussCellGradVariable import _GaussCellGradVariable
from fipy.tools import inline
from fipy.tools import numerix

class _ModCellGradVariable(_GaussCellGradVariable):
    def __init__(self, var, modIn, modPy):
        _GaussCellGradVariable.__init__(self, var)
        self.modIn = modIn
        self.modPy = modPy


    def _calcValueInline(self, N, M, ids, orientations, volumes):
        val = self._array.copy()

        inline._runIterateElementInline(self.modIn + """
            ITEM(val, i, vec) = 0.;

            int k;
            for (k = 0; k < M; k++) {
                int id = ITEM(ids, i, &k);
                ITEM(val, i, vec) += ITEM(orientations, i, &k) * ITEM(areaProj, id, vec) * ITEM(faceValues, id, NULL);
            }

            ITEM(val, i, vec) /= ITEM(volumes, i, NULL);
            ITEM(val, i, vec) = mod(ITEM(val, i, vec) * gridSpacing[vec[0]]) /  gridSpacing[vec[0]];
        """, val = val,
            ids = numerix.array(ids),
            orientations = numerix.array(orientations),
            volumes = numerix.array(volumes),
            areaProj = numerix.array(self.mesh._areaProjections),
            faceValues = numerix.array(self.var.arithmeticFaceValue),
            M = M,
            ni = N,
            gridSpacing = numerix.array(self.mesh._meshSpacing),
            shape=numerix.array(numerix.shape(val)))

        return self._makeValue(value = val)

    def _calcValueNoInline(self, N, M, ids, orientations, volumes):
        value = _GaussCellGradVariable._calcValueNoInline(self, N, M, ids, orientations, volumes)
        gridSpacing = self.mesh._meshSpacing
        return self.modPy(value * gridSpacing) / gridSpacing



from __future__ import unicode_literals
from builtins import range
__all__ = ["doInline"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

import inspect
import os
import sys

if '--inline' in [s.lower() for s in sys.argv[1:]]:
    doInline = True
else:
    doInline = 'FIPY_INLINE' in os.environ

_inlineFrameComment = 'FIPY_INLINE_COMMENT' in os.environ

def _getframeinfo(level, context=1):
    """
    Much faster alternative to `inspect.getouterframes(inspect.currentframe())[level]`
    """
    frame = inspect.currentframe()
    for l in range(level+1):
        frame = frame.f_back

    return (frame,) + inspect.getframeinfo(frame, context=context)

def _rawCodeComment(code, level=2):
    if _inlineFrameComment:
        finfo = _getframeinfo(level=level)

        # note:
        # don't use #line because it actually makes it harder
        # to find the offending code in both the C++ source and in the Python
        #line %d "%s"

        return '''
/*
    %s:%d

    %s
*/
        ''' % (finfo[1], finfo[2] - len(code.splitlines()), finfo[3])
    else:
        return ""

def _operatorVariableComment(canInline=True, level=3):
    if canInline and doInline and _inlineFrameComment:
        finfo = _getframeinfo(level=level)

        # note:
        # don't use #line because it actually makes it harder
        # to find the offending code in both the C++ source and in the Python
        #line %d "%s"

        if finfo[4] is not None:
            code = "\n".join(finfo[4])
        else:
            code = ""

        return '''
/*
%s:%d

%s
 */
        ''' % (finfo[1], finfo[2], code)
    else:
        return ""

def _runInline(code_in, converters=None, verbose=0, comment=None, **args):
    argsKeys = list(args.keys())
    dimList = ['i', 'j', 'k']

    if 'ni' in argsKeys:
        dimensions = 1
        if 'nj' in argsKeys:
            dimensions = 2
            if 'nk' in argsKeys:
                dimensions = 3
    else:
        dimensions = 0

    if dimensions == 0:
        code = """ { %s } """ % code_in
    else:
        loops = """"""
        enders = """"""
        declarations = []
        for dim in range(dimensions):
            d = dimList[dim]
            declarations.append(d)
            loops += "\t" * dim + "for(%s=0;%s<n%s;%s++) {\n" % (d, d, d, d)
            enders += "\n" + "\t" * (dimensions - dim -1) + "}"
        code = 'int ' + ','.join(declarations) + ';\n' + loops + "\t" * dimensions + code_in + enders

    if comment is None:
        comment = _rawCodeComment(code_in)

    code = "\n" + comment + "\n" + code

    import weave

    for key in list(args.keys()):
        if hasattr(args[key], 'dtype') and args[key].dtype.char == '?':
            args[key] = args[key].astype('B')

    weave.inline(code,
                 list(args.keys()),
                 local_dict=args,
                 type_converters=None, #weave.converters.blitz,
                 compiler = 'gcc',
                 force=0,
                 verbose = 0 or verbose,
                 extra_compile_args =['-O3'])

def _runIterateElementInline(code_in, converters=None, verbose=0, comment=None, **args):
    loops = """
int i;
for(i=0; i < ni; i++) {

"""

    enders = ""

    shape = args['shape']
    rank = len(shape) - 1
    for dim in range(rank):
        loops += "\t" * (dim + 1) + "for (vec[%(dim)d]=0; vec[%(dim)d] < shape[%(dim)d]; vec[%(dim)d]++) {\n" % {'dim': dim}
        enders += "\n" + "\t" * (rank - dim) + "}"

    enders += """

}
"""

    indent = "\t" * rank

    code = """
    #define ITEM(arr,i,vec) (arr[arrayIndex(arr##_array, i, vec)])

    int vec[%(rank)d];
    %(loops)s%(indent)s%(code_in)s%(enders)s

    #undef ITEM
    """ % locals()

    if comment is None:
        comment = _rawCodeComment(code_in)

    code = "\n" + comment + "\n" + code

    import weave

    for key in list(args.keys()):
        if hasattr(args[key], 'dtype') and args[key].dtype.char == '?':
            args[key] = args[key].astype('B')

    weave.inline(code,
                 list(args.keys()),
                 local_dict=args,
                 type_converters=None, #weave.converters.blitz,
                 compiler = 'gcc',
                 force=0,
                 verbose = 0 or verbose,
                 extra_compile_args =['-O3'],
                 support_code="""

// returns the index (accounting for strides) of the tensor element vec
// in position i of array
//
// array holds a tensor at each position i
// vec identifies a particular element in that tensor
static int arrayIndex(PyArrayObject* array, int i, int vec[])
{
    int index = array->strides[array->nd-1] * i;

    if (vec != NULL) {
        int j;
        for (j=0; j < array->nd - 1; j++) {
            index += array->strides[j] * vec[j];
        }
    }

    return index / array->descr->elsize;
}
                 """)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.tools import numerix
from fipy.variables.faceVariable import FaceVariable
from fipy.variables.cellVariable import CellVariable

from fipy.viewers.matplotlibViewer.matplotlib2DViewer import AbstractMatplotlib2DViewer

__all__ = ["MatplotlibStreamViewer"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class MatplotlibStreamViewer(AbstractMatplotlib2DViewer):
    """Displays a stream plot of a 2D rank-1 `CellVariable` or
    `FaceVariable` object using Matplotlib_

    One issue is that this `Viewer` relies on `scipy.interpolate.griddata`,
    which interpolates on the convex hull of the data. The results is that
    streams are plotted across any concavities in the mesh.

    Another issue is that it does not seem possible to remove the streams
    without calling `cla()`, which means that different set of streams cannot be
    overlaid.

    .. _Matplotlib: http://matplotlib.sourceforge.net/

    """

    __doc__ += AbstractMatplotlib2DViewer._test2Dvector(viewer="MatplotlibStreamViewer")
    __doc__ += AbstractMatplotlib2DViewer._test2DvectorIrregular(viewer="MatplotlibStreamViewer")

    def __init__(self, vars, title=None, log=False, limits={}, axes=None, figaspect='auto',
                 density=1, linewidth=None, color=None, cmap=None, norm=None, arrowsize=1,
                 arrowstyle='-|>', minlength=0.1,
                 **kwlimits):
        """Creates a `MatplotlibStreamViewer`.

        Parameters
        ----------
        vars : ~fipy.variables.cellVariable.CellVariable or ~fipy.variables.faceVariable.FaceVariable
            rank-1 `Variable` to display
        title : str, optional
            displayed at the top of the `Viewer` window
        log : bool, optional
            if `True`, arrow length goes at the base-10 logarithm of the magnitude
        limits : dict, optional
            a (deprecated) alternative to limit keyword arguments
        float xmin, xmax, ymin, ymax, datamin, datamax : float
            displayed range of data. Any limit set to
            a (default) value of `None` will autoscale.
        axes : ~matplotlib.axes.Axes, optional
            if not `None`, `vars` will be plotted into this Matplotlib `Axes` object
        figaspect : float, optional
            desired aspect ratio of figure. If arg is a number, use that aspect
            ratio. If arg is `auto`, the aspect ratio will be determined from
            the Variable's mesh.
        density : float or (float, float), optional
            Controls the closeness of streamlines.  When ``density = 1``,
            the domain is divided into a 30x30 grid.  *density* linearly
            scales this grid.  Each cell in the grid can have, at most, one
            traversing streamline.  For different densities in each
            direction, use a tuple (density_x, density_y).
        linewidth : array_like or ~fipy.variables.cellVariable.CellVariable or ~fipy.variables.faceVariable.FaceVariable, optional
            The width of the stream lines.  With a rank-0 `CellVariable` or
            `FaceVariable` the line width can be varied across the grid.
            The MeshVariable must have the same type and be defined on
            the same `Mesh` as *vars*.
        color : str or ~fipy.variables.cellVariable.CellVariable or ~fipy.variables.faceVariable.FaceVariable, optional
            The streamline color as a matplotlib color code or a field of
            numbers.  If given a rank-0 `CellVariable` or `FaceVariable`,
            its values are converted to colors using *cmap* and *norm*.
            The MeshVariable must have the same type and be defined on the
            same `Mesh` as *vars*.
        cmap : ~matplotlib.colors.Colormap, optional
            Colormap used to plot streamlines and arrows.  This is only
            used if *color* is a MeshVariable.
        norm : ~matplotlib.colors.Normalize, optional
            Normalize object used to scale luminance data to 0, 1.  If
            ``None``, stretch (min, max) to (0, 1).  Only necessary when
            *color* is a MeshVariable.
        arrowsize : float, optional
            Scaling factor for the arrow size.
        arrowstyle : str, optional
            Arrow style specification.
            See `~matplotlib.patches.FancyArrowPatch`.
        minlength : float, optional
              Minimum length of streamline in axes coordinates.

        """
        kwlimits.update(limits)
        AbstractMatplotlib2DViewer.__init__(self, vars=vars, title=title, axes=axes, figaspect=figaspect, **kwlimits)

        self.log = log
        self.kwargs = dict(density=density, cmap=cmap, norm=norm, arrowsize=arrowsize,
                           arrowstyle=arrowstyle, minlength=minlength)
        self.linewidth = linewidth
        self.color = color

        self._stream = None

        self._plot()

    def _getSuitableVars(self, vars):
        from fipy.meshes.mesh2D import Mesh2D
        from fipy.meshes.uniformGrid2D import UniformGrid2D

        vars = [var for var in AbstractMatplotlib2DViewer._getSuitableVars(self, vars) \
                if ((isinstance(var.mesh, Mesh2D)
                     or isinstance(var.mesh, UniformGrid2D))\
                    and (isinstance(var, FaceVariable) \
                         or isinstance(var, CellVariable)) and var.rank == 1)]
        if len(vars) == 0:
            from fipy.viewers import MeshDimensionError
            raise MeshDimensionError("The mesh must be a Mesh2D instance")
        # this viewer can only display one variable
        return [vars[0]]

    def _plot(self):
        from scipy.interpolate import griddata

        var = self.vars[0]
        mesh = var.mesh

        xmin, ymin = mesh.extents['min']
        xmax, ymax = mesh.extents['max']

        N = 100
        X = numerix.linspace(xmin, xmax, N)
        Y = numerix.linspace(ymin, ymax, N)

        grid_x, grid_y = numerix.mgrid[xmin:xmax:N*1j, ymin:ymax:N*1j]

        if isinstance(var, FaceVariable):
            C = mesh.faceCenters
        elif isinstance(var, CellVariable):
            C = mesh.cellCenters

        U = griddata(C.value.T, var.value[0],
                     (grid_x, grid_y), method='cubic')
        V = griddata(C.value.T, var.value[1],
                     (grid_x, grid_y), method='cubic')

        lw = self.linewidth
        if isinstance(lw, (FaceVariable, CellVariable)):
            lw = griddata(C.value.T, lw.value,
                          (grid_x, grid_y), method='cubic')

        color = self.color
        if isinstance(color, (FaceVariable, CellVariable)):
            color = griddata(C.value.T, color.value,
                             (grid_x, grid_y), method='cubic', fill_value=color.min())

        U = U.T
        V = V.T

        ang = numerix.arctan2(V, U)
        mag = numerix.sqrt(U**2 + V**2)

        datamin, datamax = self._autoscale(vars=(mag,),
                                           datamin=self._getLimit('datamin'),
                                           datamax=self._getLimit('datamax'))

        mag = numerix.where(mag > datamax, numerix.nan, mag)
        mag = numerix.where(mag < datamin, numerix.nan, mag)

        if self.log:
            mag = numerix.log10(mag)

        U = mag * numerix.cos(ang)
        V = mag * numerix.sin(ang)

#         if self._stream is not None:
#             # the following doesn't work, nor does it help to `add_collection` first
#             # self._stream.arrows.remove()
#             self._stream.lines.remove()

        self.axes.cla()
        self._stream = self.axes.streamplot(X, Y, U, V, linewidth=lw, color=color, **self.kwargs)

        self.axes.set_xlim(xmin=self._getLimit('xmin'),
                           xmax=self._getLimit('xmax'))
        self.axes.set_ylim(ymin=self._getLimit('ymin'),
                           ymax=self._getLimit('ymax'))

if __name__ == "__main__":
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.variables.cellToFaceVariable import _CellToFaceVariable
from fipy.tools import inline
from fipy.tools import numerix

class _LevelSetDiffusionVariable(_CellToFaceVariable):
    r"""
    This variable sets it's face value to zero if either of the
    surrounding cell values are zero else it uses the value of the
    diffusion coefficient. The diffusion coefficient is given by,

    .. math::

        D = \begin{cases}
            D_c & \text{when $\phi > 0$} \\
            0  & \text{when $\phi \le 0$}

    Here is a simple 1D test case:

    >>> from fipy.meshes import Grid2D
    >>> mesh = Grid2D(dx = 1., nx = 3)
    >>> from fipy.variables.cellVariable import CellVariable
    >>> var = CellVariable(mesh = mesh, value = (-1, 1, 1))
    >>> from fipy.variables.faceVariable import FaceVariable
    >>> answer = FaceVariable(mesh=mesh, value=(0, 1, 1, 0, 1, 1, 0, 0, 1, 1))
    >>> print(_LevelSetDiffusionVariable(var, 1).allclose(answer))
    True
    """
    def __init__(self, distanceVariable = None, diffusionCoeff = None):
        """
        Creates a `_LevelSetDiffusionVariable`.

        Parameters
        ----------
        distanceVariable : ~fipy.variables.distanceVariable.DistanceVariable
        diffusionCoeff : float or ~fipy.variables.faceVariable.FaceVariable

        """
        _CellToFaceVariable.__init__(self, distanceVariable)
        self.diffusionCoeff = diffusionCoeff

    if inline.doInline:
        def _calcValue_(self, alpha, id1, id2):
            val = self._array.copy()

            inline._runInline("""
                int ID1 = id1[i];
                int ID2 = id2[i];
                double	cell1 = var[ID1];
                double	cell2 = var[ID2];

                if (cell1 < 0 || cell2 < 0) {
                    val[i] = 0;
                } else {
                    val[i] = diffusionCoeff;
                }
            """,
            var = numerix.array(self.var),
            val = val,
            id1 = id1, id2 = id2,
            diffusionCoeff = self.diffusionCoeff,
            ni = self.mesh.numberOfFaces
            )

            return self._makeValue(value = val)
    else:
        def _calcValue_(self, alpha, id1, id2):
            distance = numerix.array(self.var)
            cell1 = numerix.take(distance, id1)
            cell2 = numerix.take(distance, id2)

            return numerix.where(numerix.logical_or(cell1 < 0, cell2 < 0),
                                 0,
                                 self.diffusionCoeff)

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.tools import parallelComm
from fipy.tools import numerix

__all__ = ["Grid3D", "Grid2D", "Grid1D", "CylindricalGrid2D", "CylindricalGrid1D", "SphericalGrid1D"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

def _dnl(dx, nx, Lx):
    """
    Initialize arguments for grid classes based on an over determined set
    of initial arguments.  The order of precedence is `nx` then `Lx` then
    `dx`.  i.e. If `Lx` is specified the length of
    the domain is always `Lx` regardless of `dx`.

    >>> print(_dnl(None, None, None))
    (None, 1)
    >>> print(_dnl(None, 1.1, 2.))
    (2.0, 1)
    >>> print(_dnl(3., None, 7.5))
    (3.75, 2)
    >>> print("(%1.1f, %i)" % _dnl(2.2, 4, None))
    (2.2, 4)
    >>> print(_dnl(1., 6, 15.))
    (2.5, 6)

    Parameters
    ----------
    dx : float
        Grid spacing
    nx : int
        Number of cells
    Lx : float
        The domain length
    """
    if Lx is None:
        if nx is None:
            nx = 1
    else:
        if nx is None:
            nx = Lx // dx or 1
        dx = Lx / int(nx)

    return dx, int(nx)

def Grid3D(dx=1., dy=1., dz=1.,
           nx=None, ny=None, nz=None,
           Lx=None, Ly=None, Lz=None,
           overlap=2, communicator=parallelComm):

    r""" Factory function to select between `UniformGrid3D` and
    `NonUniformGrid3D`. If `L{x,y,z}` is specified, the length of the domain
    is always `L{x,y,z}` regardless of `d{x,y,z}`, unless `d{x,y,z}` is a
    list of spacings, in which case `L{x,y,z}` will be the sum of
    `d{x,y,z}` and `n{x,y,z}` will be the count of `d{x,y,z}`.

    Parameters
    ----------
    dx : float
        Grid spacing in the horizontal direction
    dy : float
        Grid spacing in the vertical direction
    dz : float
        Grid spacing in the depth direction
    nx : int
        Number of cells in the horizontal direction
    ny : int
        Number of cells in the vertical direction
    nz : int
        Number of cells in the depth direction
    Lx : float
        Domain length in the horizontal direction
    Ly : float
        Domain length in the vertical direction
    Lz : float
        Domain length in the depth direction
    overlap : int
        Number of overlapping cells for parallel simulations.  Generally 2
        is adequate.  Higher order equations or discretizations require
        more.
    communicator : ~fipy.tools.comms.commWrapper.CommWrapper
        Generally, `fipy.tools.serialComm` or `fipy.tools.parallelComm`.
        Select `~fipy.tools.serialComm` to create a serial mesh when
        running in parallel; mostly used for test purposes.
    """

    if numerix.getShape(dx) == () \
      and numerix.getShape(dy) == () \
      and numerix.getShape(dz) == ():

        dx, nx = _dnl(dx, nx, Lx)
        dy, ny = _dnl(dy, ny, Ly)
        dz, nz = _dnl(dz, nz, Lz)
        from fipy.meshes.uniformGrid3D import UniformGrid3D
        return UniformGrid3D(dx = dx, dy = dy, dz = dz,
                             nx = nx or 1, ny = ny or 1, nz = nz or 1,
                             overlap=overlap, communicator=communicator)
    else:
        from fipy.meshes.nonUniformGrid3D import NonUniformGrid3D
        return NonUniformGrid3D(dx = dx, dy = dy, dz = dz, nx = nx, ny = ny, nz = nz,
                                overlap=overlap, communicator=communicator)

def Grid2D(dx=1., dy=1., nx=None, ny=None, Lx=None, Ly=None, overlap=2, communicator=parallelComm):
    r""" Factory function to select between `UniformGrid2D` and
    `NonUniformGrid2D`. If `L{x,y}` is specified, the length of the domain is
    always `L{x,y}` regardless of `d{x,y}`, unless `d{x,y}` is a list of
    spacings, in which case `L{x,y}` will be the sum of `d{x,y}` and
    `n{x,y}` will be the count of `d{x,y}`.

    >>> print(Grid2D(Lx=3., nx=2).dx)
    1.5

    Parameters
    ----------
    dx : float
        Grid spacing in the horizontal direction
    dy : float
        Grid spacing in the vertical direction
    nx : int
        Number of cells in the horizontal direction
    ny : int
        Number of cells in the vertical direction
    Lx : float
        Domain length in the horizontal direction
    Ly : float
        Domain length in the vertical direction
    overlap : int
        Number of overlapping cells for parallel simulations.  Generally 2
        is adequate.  Higher order equations or discretizations require
        more.
    communicator : ~fipy.tools.comms.commWrapper.CommWrapper
        Generally, `fipy.tools.serialComm` or `fipy.tools.parallelComm`.
        Select `~fipy.tools.serialComm` to create a serial mesh when
        running in parallel; mostly used for test purposes.
    """

    if numerix.getShape(dx) == () and numerix.getShape(dy) == ():

        dx, nx = _dnl(dx, nx, Lx)
        dy, ny = _dnl(dy, ny, Ly)

        from fipy.meshes.uniformGrid2D import UniformGrid2D
        return UniformGrid2D(dx=dx, dy=dy,
                             nx=nx, ny=ny,
                             overlap=overlap,
                             communicator=communicator)
    else:
        from fipy.meshes.nonUniformGrid2D import NonUniformGrid2D
        return NonUniformGrid2D(dx=dx, dy=dy, nx=nx, ny=ny, overlap=overlap, communicator=communicator)

def Grid1D(dx=1., nx=None, Lx=None, overlap=2, communicator=parallelComm):
    r""" Factory function to select between `UniformGrid1D` and
    `NonUniformGrid1D`. If `Lx` is specified the length of the domain is
    always `Lx` regardless of `dx`, unless `dx` is a list of spacings, in
    which case `Lx` will be the sum of `dx` and `nx` will be the count of
    `dx`.

    Parameters
    ----------
    dx : float
        Grid spacing in the horizontal direction
    nx : int
        Number of cells in the horizontal direction
    Lx : float
        Domain length in the horizontal direction
    overlap : int
        Number of overlapping cells for parallel simulations.  Generally 2
        is adequate.  Higher order equations or discretizations require
        more.
    communicator : ~fipy.tools.comms.commWrapper.CommWrapper
        Generally, `fipy.tools.serialComm` or `fipy.tools.parallelComm`.
        Select `~fipy.tools.serialComm` to create a serial mesh when
        running in parallel; mostly used for test purposes.
    """

    if numerix.getShape(dx) == ():
        dx, nx = _dnl(dx, nx, Lx)
        from fipy.meshes.uniformGrid1D import UniformGrid1D
        return UniformGrid1D(dx=dx, nx=nx, overlap=overlap, communicator=communicator)
    else:
        from fipy.meshes.nonUniformGrid1D import NonUniformGrid1D
        return NonUniformGrid1D(dx=dx, nx=nx, overlap=overlap, communicator=communicator)

def CylindricalGrid2D(dr=None, dz=None,
                      nr=None, nz=None,
                      Lr=None, Lz=None,
                      dx=1., dy=1.,
                      nx=None, ny=None,
                      Lx=None, Ly=None,
                      origin=((0,), (0,)),
                      overlap=2,
                      communicator=parallelComm):

    r""" Factory function to select between `CylindricalUniformGrid2D` and
    `CylindricalNonUniformGrid2D`. If `Lr` is specified the length of the
    domain is always `Lr` regardless of `dr`, unless `dr` is a list of
    spacings, in which case `Lr` will be the sum of `dr`.

    Parameters
    ----------
    dr : float
        Grid spacing in the radial direction. Alternative: `dx`.
    dz : float
        grid spacing in the vertical direction. Alternative: `dy`.
    nr : int
        Number of cells in the radial direction. Alternative: `nx`.
    nz : int
        Number of cells in the vertical direction. Alternative: `ny`.
    Lr : float
        Domain length in the radial direction. Alternative: `Lx`.
    Lz : float
        Domain length in the vertical direction. Alternative: `Ly`.
    overlap : int
        the number of overlapping cells for parallel
        simulations. Generally 2 is adequate. Higher order equations or
        discretizations require more.
    communicator : ~fipy.tools.comms.commWrapper.CommWrapper
        Generally, `fipy.tools.serialComm` or `fipy.tools.parallelComm`.
        Select `~fipy.tools.serialComm` to create a serial mesh when
        running in parallel; mostly used for test purposes.
    """
    if dr is not None:
        dx = dr

    if dz is not None:
        dy = dz


    nx = nr or nx
    ny = nz or ny

    Lx = Lr or Lx
    Ly = Lz or Ly

    if numerix.getShape(dx) == () and numerix.getShape(dy) == ():

        dx, nx = _dnl(dx, nx, Lx)
        dy, ny = _dnl(dy, ny, Ly)
        from fipy.meshes.cylindricalUniformGrid2D import CylindricalUniformGrid2D
        return CylindricalUniformGrid2D(dx=dx, dy=dy, nx=nx or 1, ny=ny or 1, origin=origin, overlap=overlap, communicator=communicator)
    else:
        from fipy.meshes.cylindricalNonUniformGrid2D import CylindricalNonUniformGrid2D
        return CylindricalNonUniformGrid2D(dx=dx, dy=dy, nx=nx, ny=ny, origin=origin, overlap=overlap, communicator=communicator)

def CylindricalGrid1D(dr=None, nr=None, Lr=None,
                      dx=1., nx=None, Lx=None,
                      origin=(0,), overlap=2, communicator=parallelComm):

    r""" Factory function to select between `CylindricalUniformGrid1D` and
    `CylindricalNonUniformGrid1D`. If `Lr` is specified the length of the
    domain is always `Lr` regardless of `dr`, unless `dr` is a list of
    spacings, in which case `Lr` will be the sum of `dr`.

    Parameters
    ----------
    dr : float
        Grid spacing in the radial direction. Alternative: `dx`.
    nr : int
        Number of cells in the radial direction. Alternative: `nx`.
    Lr : float
        Domain length in the radial direction. Alternative: `Lx`.
    overlap : int
        the number of overlapping cells for parallel
        simulations. Generally 2 is adequate. Higher order equations or
        discretizations require more.
    communicator : ~fipy.tools.comms.commWrapper.CommWrapper
        Generally, `fipy.tools.serialComm` or `fipy.tools.parallelComm`.
        Select `~fipy.tools.serialComm` to create a serial mesh when
        running in parallel; mostly used for test purposes.
    """

    if dr is not None:
        dx = dr

    nx = nr or nx
    Lx = Lr or Lx

    if numerix.getShape(dx) == ():
        dx, nx = _dnl(dx, nx, Lx)
        from fipy.meshes.cylindricalUniformGrid1D import CylindricalUniformGrid1D
        return CylindricalUniformGrid1D(dx=dx, nx=nx or 1, origin=origin, overlap=overlap, communicator=parallelComm)
    else:
        from fipy.meshes.cylindricalNonUniformGrid1D import CylindricalNonUniformGrid1D
        return CylindricalNonUniformGrid1D(dx=dx, nx=nx, origin=origin, overlap=overlap, communicator=parallelComm)

def SphericalGrid1D(dr=None, nr=None, Lr=None,
                    dx=1., nx=None, Lx=None,
                    origin=(0,), overlap=2, communicator=parallelComm):

    r""" Factory function to select between `SphericalUniformGrid1D` and
    `SphericalNonUniformGrid1D`. If `Lr` is specified the length of the
    domain is always `Lr` regardless of `dr`, unless `dr` is a list of
    spacings, in which case `Lr` will be the sum of `dr`.

    Parameters
    ----------
    dr : float
        Grid spacing in the radial direction. Alternative: `dx`.
    nr : int
        Number of cells in the radial direction. Alternative: `nx`.
    Lr : float
        Domain length in the radial direction. Alternative: `Lx`.
    overlap : int
        the number of overlapping cells for parallel
        simulations. Generally 2 is adequate. Higher order equations or
        discretizations require more.
    communicator : ~fipy.tools.comms.commWrapper.CommWrapper
        Generally, `fipy.tools.serialComm` or `fipy.tools.parallelComm`.
        Select `~fipy.tools.serialComm` to create a serial mesh when
        running in parallel; mostly used for test purposes.
    """

    if dr is not None:
        dx = dr

    nx = nr or nx
    Lx = Lr or Lx

    if numerix.getShape(dx) == ():
        dx, nx = _dnl(dx, nx, Lx)
        from fipy.meshes.sphericalUniformGrid1D import SphericalUniformGrid1D
        return SphericalUniformGrid1D(dx=dx, nx=nx or 1, origin=origin, overlap=overlap, communicator=parallelComm)
    else:
        from fipy.meshes.sphericalNonUniformGrid1D import SphericalNonUniformGrid1D
        return SphericalNonUniformGrid1D(dx=dx, nx=nx, origin=origin, overlap=overlap, communicator=parallelComm)
    
def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()




from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

import optparse
import sys, os

__all__ = ["parse"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

def parse(larg, action = None, type = None, default = None):
    """
    This is a wrapper function for the python `optparse` module.
    Unfortunately `optparse` does not allow command line arguments to
    be ignored. See the documentation for `optparse` for more
    details. Returns the argument value.

    Parameters
    ----------
    larg : str
        Argument to be parsed.
    action : {`'store'`, `'store_true'`, `'store_false'`, `'store_const'`, `'append'`, `'count'`, `'callback'`}
        Basic type of action to be taken when this argument is encountered at the command line.
        See https://docs.python.org/2/library/argparse.html#action
    type : type
        Type to which the command-line argument should be converted
    default
        Value produced if the argument is absent from the command line
    """
    sarg = None
    tmpparser = optparse.OptionParser(option_list = [
        optparse.make_option(sarg, larg, action = action, type = type, dest = 'dest', default = default)],
                                      conflict_handler = 'resolve')

##    optparse.make_option('-e', '--numberOfElements', action = 'store', type = 'int', dest = 'Nele', default = numberOfElements),
##    optparse.make_option('-n', '--numberOfSteps', action = 'store', type = 'int', dest = 'steps', default = numberOfSteps),
##    optparse.make_option('-p', '--quiet', action = 'store_true', dest = 'quiet', default = False),
##    optparse.make_option('-i', '--inline', action = 'store_true', dest = 'inline', default = False)])

    sysargs = []
    for arg in sys.argv:
        if larg in arg:
            sysargs.append(arg)

    (options, args) = tmpparser.parse_args(sysargs)

    return options.dest

def _parseSolver():
    args = [s.lower() for s in sys.argv[1:]]
    # any command-line specified solver takes precedence over environment variables
    if '--no-pysparse' in args:
        return "no-pysparse"
    elif '--trilinos' in args:
        return "trilinos"
    elif '--pysparse' in args:
        return "pysparse"
    elif '--pyamg' in args:
        return 'pyamg'
    elif '--scipy' in args:
        return 'scipy'
    elif '--pyamgx' in args:
        return 'pyamgx'
    elif '--petsc' in args:
        return 'petsc'
    elif 'FIPY_SOLVERS' in os.environ:
        return os.environ['FIPY_SOLVERS'].lower()
    else:
        return None



from __future__ import print_function
from __future__ import unicode_literals
from builtins import range
__all__ = ["PRINT"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

def PRINT(label, arg="", stall=True):
    import sys
    from fipy import parallelComm
    import time

    for procID in range(parallelComm.Nproc):
        if procID == parallelComm.procID:
            print(parallelComm.procID, label, arg, file=sys.stderr)
        sys.stderr.flush()
        if stall:
            time.sleep(0.1)
            parallelComm.Barrier()

    if stall:
        parallelComm.Barrier()



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from PyTrilinos import AztecOO

from fipy.solvers.trilinos.trilinosAztecOOSolver import TrilinosAztecOOSolver
from fipy.solvers.trilinos.preconditioners.jacobiPreconditioner import JacobiPreconditioner

__all__ = ["LinearBicgstabSolver"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class LinearBicgstabSolver(TrilinosAztecOOSolver):

    """
    The `LinearBicgstabSolver` is an interface to the biconjugate gradient
    stabilized solver in Trilinos, using the `JacobiPreconditioner` by default.

    """

    def __init__(self, tolerance=1e-10, iterations=1000, precon=JacobiPreconditioner()):
        """
        Parameters
        ----------
        tolerance : float
            Required error tolerance.
        iterations : int
            Maximum number of iterative steps to perform.
        precon : ~fipy.solvers.trilinos.preconditioners.preconditioner.Preconditioner
        """
        TrilinosAztecOOSolver.__init__(self, tolerance=tolerance,
                                       iterations=iterations, precon=precon)
        self.solver = AztecOO.AZ_bicgstab



"""
1D Mesh
"""
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.tools import numerix
from fipy.tools.dimensions.physicalField import PhysicalField
from fipy.tools import parallelComm

from fipy.meshes.nonUniformGrid1D import NonUniformGrid1D

__all__ = ["CylindricalNonUniformGrid1D"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class CylindricalNonUniformGrid1D(NonUniformGrid1D):
    """
    Creates a 1D cylindrical grid mesh.

        >>> mesh = CylindricalNonUniformGrid1D(nx = 3)
        >>> print(mesh.cellCenters)
        [[ 0.5  1.5  2.5]]

        >>> mesh = CylindricalNonUniformGrid1D(dx = (1, 2, 3))
        >>> print(mesh.cellCenters)
        [[ 0.5  2.   4.5]]

        >>> print(numerix.allclose(mesh.cellVolumes, (0.5, 4., 13.5))) # doctest: +PROCESSOR_0
        True

        >>> mesh = CylindricalNonUniformGrid1D(nx = 2, dx = (1, 2, 3))
        Traceback (most recent call last):
        ...
        IndexError: nx != len(dx)

        >>> mesh = CylindricalNonUniformGrid1D(nx=2, dx=(1., 2.)) + ((1.,),)
        >>> print(mesh.cellCenters)
        [[ 1.5  3. ]]
        >>> print(numerix.allclose(mesh.cellVolumes, (1.5, 6))) # doctest: +PROCESSOR_0
        True

    """
    def __init__(self, dx=1., nx=None, origin=(0,), overlap=2, communicator=parallelComm, *args, **kwargs):
        scale = PhysicalField(value=1, unit=PhysicalField(value=dx).unit)
        self.origin = PhysicalField(value=origin)
        self.origin /= scale

        super(CylindricalNonUniformGrid1D, self).__init__(dx=dx,
                                                          nx=nx,
                                                          overlap=overlap,
                                                          communicator=communicator,
                                                          *args,
                                                          **kwargs)

        self.vertexCoords += origin
        self.args['origin'] = origin

    def _calcFaceCenters(self):
        faceCenters = super(CylindricalNonUniformGrid1D, self)._calcFaceCenters()
        return faceCenters + self.origin

    def _calcFaceAreas(self):
        return self._calcFaceCenters()[0]

    def _calcCellVolumes(self):
        return super(CylindricalNonUniformGrid1D, self)._calcCellVolumes() / 2.

    def _translate(self, vector):
        return CylindricalNonUniformGrid1D(dx=self.args['dx'], nx=self.args['nx'],
                                           origin=numerix.array(self.args['origin']) + vector,
                                           overlap=self.args['overlap'])

    def __mul__(self, factor):
        return CylindricalNonUniformGrid1D(dx=self.args['dx'] * factor, nx=self.args['nx'],
                                           origin=numerix.array(self.args['origin']) * factor,
                                           overlap=self.args['overlap'])

    def _test(self):
        """
        These tests are not useful as documentation, but are here to ensure
        everything works as expected. Fixed a bug where the following throws
        an error on solve() when `nx` is a float.

            >>> from fipy import CellVariable, DiffusionTerm
            >>> mesh = CylindricalNonUniformGrid1D(nx=3., dx=(1., 2., 3.))
            >>> var = CellVariable(mesh=mesh)
            >>> DiffusionTerm().solve(var)

        This test is for https://github.com/usnistgov/fipy/issues/372. Cell
        volumes were being returned as `binOps` rather than arrays.

            >>> m = CylindricalNonUniformGrid1D(dx=(1., 2., 3., 4.), nx=4)
            >>> print(isinstance(m.cellVolumes, numerix.ndarray))
            True
            >>> print(isinstance(m._faceAreas, numerix.ndarray))
            True

        If the above types aren't correct, the divergence operator's value can be a `binOp`

            >>> print(isinstance(CellVariable(mesh=m).arithmeticFaceValue.divergence.value, numerix.ndarray))
            True

        """

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()




from __future__ import print_function
from __future__ import unicode_literals
from builtins import object
__docformat__ = 'restructuredtext'

__all__ = ["AbstractViewer"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

import sys

class AbstractViewer(object):
    """
    .. attention:: This class is abstract. Always create one of its subclasses.
    """
    def __init__(self, vars, title=None, **kwlimits):
        """Create a `AbstractViewer` object.

        Parameters
        ----------
        vars : ~fipy.variables.cellVariable.CellVariable or list
            the `CellVariable` objects to display.
        title : str, optional
            displayed at the top of the `Viewer` window
        float xmin, xmax, ymin, ymax, zmin, zmax, datamin, datamax : float, optional
            displayed range of data. Any limit set to
            a (default) value of `None` will autoscale.
        """
        if self.__class__ is AbstractViewer:
            raise NotImplementedError("can't instantiate abstract base class")

        self.vars = self._getSuitableVars(vars)

        self.limits = kwlimits

        if title is None:
            if len(self.vars) == 1:
                title = self.vars[0].name
            else:
                title = ''

        self.title = title

    def _getSuitableVars(self, vars):
        if type(vars) not in [type([]), type(())]:
            vars = [vars]

        return [var for var in vars]

    def setLimits(self, limits={}, **kwlimits):
        """
        Update the limits.

        Parameters
        ----------
        limits : dict, optional
            a (deprecated) alternative to limit keyword arguments
        float xmin, xmax, ymin, ymax, zmin, zmax, datamin, datamax : float, optional
            displayed range of data. A 1D `Viewer` will only use `xmin` and
            `xmax`, a 2D viewer will also use `ymin` and `ymax`, and so on. All
            viewers will use `datamin` and `datamax`. Any limit set to a
            (default) value of `None` will autoscale.

        """
        self.limits.update(limits)
        self.limits.update(kwlimits)

    def _getLimit(self, keys, default=None):
        """
        Return the limit associated with the first available key in `keys`

        Parameters
        ----------
        keys : str or :obj:`tuple` of :obj:`str`
            dictionary keys that identify the limits of interest

        Returns
        -------
        float or None
            the value of the limit
        """
        if not (isinstance(keys, list) or isinstance(keys, tuple)):
            keys = (keys,)
        limit = default
        for key in keys:
            limit = self.limits.get(key, limit)
            if limit is not None:
                break

        return limit

    def plot(self, filename=None):
        """
        Update the display of the viewed variables.

        Parameters
        ----------
        filename : str
            If not `None`, the name of a file to save the image into.
        """

        raise NotImplementedError

    def plotMesh(self, filename=None):
        """
        Display a representation of the mesh

        Parameters
        ----------
        filename : str
            If not `None`, the name of a file to save the image into.
        """
        pass

    def _autoscale(self, vars, datamin=None, datamax=None):
        from fipy.tools import numerix

        if datamin is None:
            datamin = 1e300
            for var in vars:
                datamin = min(datamin, numerix.nanmin(var))

        if datamax is None:
            from fipy.tools import numerix
            datamax = -1e300
            for var in vars:
                datamax = max(datamax, numerix.nanmax(var))

        return datamin, datamax

    def _validFileExtensions(self):
        return []

    _saved_stdout = sys.stdout

    @staticmethod
    def _serial_doctest_raw_input(prompt):
        """Replacement for `raw_input()` that works in doctests
        """
        AbstractViewer._saved_stdout.write("\n")
        AbstractViewer._saved_stdout.write(prompt)
        AbstractViewer._saved_stdout.flush()
        return sys.stdin.readline()

    @staticmethod
    def _doctest_raw_input(prompt):
        """Replacement for `raw_input()` that works in doctests

        This routine attempts to be savvy about running in parallel.
        """
        try:
            from fipy.tools import parallelComm
            parallelComm.Barrier()
            AbstractViewer._saved_stdout.flush()
            if parallelComm.procID == 0:
                txt = AbstractViewer._serial_doctest_raw_input(prompt)
            else:
                txt = ""
            parallelComm.Barrier()
        except ImportError:
            txt = AbstractViewer._serial_doctest_raw_input(prompt)
        return txt


    def _promptForOpinion(self, prompt="Describe any problems with this figure or hit Return: "):
        # This method is usually invoked from a test, which can have a weird
        # state; In particular, it may have a special `raw_input` to prevent user
        # interaction during the test.

        opinion = self._doctest_raw_input(self.__class__.__name__ + ": " + prompt).strip()
        if len(opinion) > 0:
            extensions = ", ".join(self._validFileExtensions())
            if len(extensions) > 0:
                extensions = " (%s)" % extensions
            snapshot = self._doctest_raw_input("Enter a filename%s to save a snapshot (leave blank to skip): " % extensions).strip()
            self.plot(snapshot)
            print(opinion)


    def _test1D(**kwargs):
        s = """
            >>> from fipy import *
            >>> mesh = Grid1D(nx=100)
            >>> x, = mesh.cellCenters
            >>> xVar = CellVariable(mesh=mesh, name="x", value=x)
            >>> k = Variable(name="k", value=0.)
            >>> viewer = VIEWERSUBSTITUTION(vars=(numerix.sin(k * xVar), numerix.cos(k * xVar / numerix.pi)),
            ...                 limits={'xmin': 10, 'xmax': 90},
            ...                 datamin=-0.9, datamax=2.0,
            ...                 title="VIEWERSUBSTITUTION test")
            >>> for kval in numerix.arange(0, 0.3, 0.03):
            ...     k.setValue(kval)
            ...     viewer.plot()
            >>> viewer._promptForOpinion()
        """
        s = s.replace("VIEWERSUBSTITUTION", kwargs["viewer"])
        return s
    _test1D = staticmethod(_test1D)

    def _test2Dbase(**kwargs):
        s = """
            >>> from fipy import *
            >>> mesh = MESHSUBSTITUTION
            >>> x, y = mesh.cellCenters
            >>> xyVar = CellVariable(mesh=mesh, name="x y", value=x * y)
            >>> k = Variable(name="k", value=0.)
            >>> viewer = VIEWERSUBSTITUTION(vars=numerix.sin(k * xyVar),
            ...                 limits={'ymin': 0.1, 'ymax': 0.9},
            ...                 datamin=-0.9, datamax=2.0,
            ...                 title="VIEWERSUBSTITUTION test")
            >>> from builtins import range
            >>> for kval in range(10):
            ...     k.setValue(kval)
            ...     viewer.plot()
            >>> viewer._promptForOpinion()
        """
        s = s.replace("VIEWERSUBSTITUTION", kwargs["viewer"])
        s = s.replace("MESHSUBSTITUTION", kwargs["mesh"])
        return s
    _test2Dbase = staticmethod(_test2Dbase)

    def _test2D(**kwargs):
        return AbstractViewer._test2Dbase(mesh="Grid2D(nx=50, ny=100, dx=0.1, dy=0.01)",
                                      **kwargs)
    _test2D = staticmethod(_test2D)

    def _test2Dirregular(**kwargs):
        """"""
        return AbstractViewer._test2Dbase(mesh="""(Grid2D(nx=5, ny=10, dx=0.1, dy=0.1)
            ...         + (Tri2D(nx=5, ny=5, dx=0.1, dy=0.1)
            ...          + ((0.5,), (0.2,))))""", **kwargs)
    _test2Dirregular = staticmethod(_test2Dirregular)

    def _test2DvectorBase(**kwargs):
        s = """
            >>> from fipy import *
            >>> mesh = MESHSUBSTITUTION
            >>> x, y = mesh.cellCenters
            >>> xyVar = CellVariable(mesh=mesh, name="x y", value=x * y)
            >>> k = Variable(name="k", value=1.)
            >>> viewer = VIEWERSUBSTITUTION(vars=numerix.sin(k * xyVar).grad,
            ...                 title="VIEWERSUBSTITUTION test")
            >>> for kval in numerix.arange(1, 10):
            ...     k.setValue(kval)
            ...     viewer.plot()
            >>> viewer._promptForOpinion()

            >>> viewer = VIEWERSUBSTITUTION(vars=numerix.sin(k * xyVar).faceGrad,
            ...                 title="VIEWERSUBSTITUTION test")
            >>> for kval in numerix.arange(1, 10):
            ...     k.setValue(kval)
            ...     viewer.plot()
            >>> viewer._promptForOpinion()
        """
        s = s.replace("VIEWERSUBSTITUTION", kwargs["viewer"])
        s = s.replace("MESHSUBSTITUTION", kwargs["mesh"])
        return s

    _test2DvectorBase = staticmethod(_test2DvectorBase)

    def _test2Dvector(**kwargs):
        return AbstractViewer._test2DvectorBase(mesh="Grid2D(nx=50, ny=100, dx=0.1, dy=0.01)",
                                            **kwargs)
    _test2Dvector = staticmethod(_test2Dvector)

    def _test2DvectorIrregular(**kwargs):
        """"""
        return AbstractViewer._test2DvectorBase(mesh="""(Grid2D(nx=5, ny=10, dx=0.1, dy=0.1)
            ...         + (Tri2D(nx=5, ny=5, dx=0.1, dy=0.1)
            ...          + ((0.5,), (0.2,))))""", **kwargs)
    _test2DvectorIrregular = staticmethod(_test2DvectorIrregular)


    def _test3D(**kwargs):
        s = """
            >>> from fipy import *
            >>> mesh = Grid3D(nx=50, ny=100, nz=10, dx=0.1, dy=0.01, dz=0.1)
            >>> x, y, z = mesh.cellCenters
            >>> xyzVar = CellVariable(mesh=mesh, name=r"x y z", value=x * y * z)
            >>> k = Variable(name="k", value=0.)
            >>> viewer = VIEWERSUBSTITUTION(vars=numerix.sin(k * xyzVar),
            ...                     limits={'ymin': 0.1, 'ymax': 0.9},
            ...                     datamin=-0.9, datamax=2.0,
            ...                     title="VIEWERSUBSTITUTION test")
            >>> from builtins import range
            >>> for kval in range(10):
            ...     k.setValue(kval)
            ...     viewer.plot()
            >>> viewer._promptForOpinion()
        """
        s = s.replace("VIEWERSUBSTITUTION", kwargs["viewer"])
        return s

    _test3D = staticmethod(_test3D)



from __future__ import unicode_literals
from fipy.meshes.nonUniformGrid1D import NonUniformGrid1D as Grid1D
import warnings
warnings.warn("Grid1D has been deprecated use NonUniformGrid1D instead.", stacklevel=3)



from __future__ import unicode_literals
from fipy.meshes.nonUniformGrid2D import NonUniformGrid2D as Grid2D
import warnings
warnings.warn("Grid2D has been deprecated use NonUniformGrid2D instead.", stacklevel=3)



from __future__ import unicode_literals
from fipy.boundaryConditions.constraint import *
from fipy.boundaryConditions.fixedFlux import *
from fipy.boundaryConditions.fixedValue import *
from fipy.boundaryConditions.nthOrderBoundaryCondition import *

__all__ = []
__all__.extend(constraint.__all__)
__all__.extend(fixedFlux.__all__)
__all__.extend(fixedValue.__all__)
__all__.extend(nthOrderBoundaryCondition.__all__)



from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from fipy.variables.cellVariable import CellVariable
from fipy.meshes import Grid1D
from fipy.tools import numerix

__all__ = ["HistogramVariable"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class HistogramVariable(CellVariable):
    def __init__(self, distribution, dx = 1., nx = None, offset = 0.):
        r"""
        Produces a histogram of the values of the supplied distribution.

        Parameters
        ----------
        distribution : array_like or ~fipy.variables.Variable
            The collection of values to sample.
        dx : float
            The bin size
        nx : int
            The number of bins
        offset : float
            The position of the first bin
        """
        CellVariable.__init__(self, mesh = Grid1D(dx = dx, nx = nx) + (offset,))
        self.distribution = self._requires(distribution)

    def _calcValue(self):
        l = len(self.distribution)
        bins = self.mesh.cellCenters[0]
        n = numerix.searchsorted(numerix.sort(self.distribution), bins)
        n = numerix.concatenate([n, [l]])
        dx = bins[1:] - bins[:-1]
        return (n[1:] - n[:-1]) / numerix.concatenate([dx, [dx[-1]]]) / float(l)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = []

from fipy.tools import numerix
from fipy.tools import inline
from fipy.variables.arithmeticCellToFaceVariable import _ArithmeticCellToFaceVariable

class _ModCellToFaceVariable(_ArithmeticCellToFaceVariable):
    def __init__(self, var, modIn):
        _ArithmeticCellToFaceVariable.__init__(self, var)
        self.modIn = modIn

    if inline.doInline:
        def  _calcValue_(self, alpha, id1, id2):
            val = self._array.copy()

            inline._runInline(self.modIn + """
            int ID1 = id1[i];
            int ID2 = id2[i];
            double cell2 = var[ID2];
            val[i] = mod(cell2 - var[ID1]) * alpha[i] + var[ID1];
            """, var = self.var.numericValue,
                val = val,
                alpha = alpha,
                id1 = id1, id2 = id2,
                ni = self.mesh.numberOfFaces)

            return self._makeValue(value = val)



from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

from PyTrilinos import IFPACK

from fipy.solvers.trilinos.preconditioners.preconditioner import Preconditioner

__all__ = ["ICPreconditioner"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class ICPreconditioner(Preconditioner):
    """
    Incomplete Cholesky Preconditioner from IFPACK for Trilinos Solvers.

    """

    def _applyToSolver(self, solver, matrix):
        Factory = IFPACK.Factory()
        Prec = Factory.Create(text_to_native_str("IC"), matrix)
        Prec.Initialize()
        Prec.Compute()
        solver.SetPrecOperator(Prec)



from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import range

from fipy.solvers import serialComm, parallelComm
serial, parallel = serialComm, parallelComm

from fipy.tests.doctestPlus import register_skipper

register_skipper(flag="SERIAL",
                 test=lambda: parallelComm.Nproc == 1,
                 why="more than one processor found",
                 skipWarning=False)

register_skipper(flag="PARALLEL",
                 test=lambda: parallelComm.Nproc > 1,
                 why="only one processor found",
                 skipWarning=False)

register_skipper(flag="PROCESSOR_0",
                 test=lambda: parallelComm.procID == 0,
                 why="not running on processor 0",
                 skipWarning=False)

register_skipper(flag="PROCESSOR_NOT_0",
                 test=lambda: parallelComm.procID > 0,
                 why="running on processor 0",
                 skipWarning=False)

for M in (2, 3):
    for N in range(M):
        register_skipper(flag="PROCESSOR_%d_OF_%d" % (N, M),
                         test=lambda N=N, M=M: parallelComm.procID == N and parallelComm.Nproc == M,
                         why="not running on processor %d of %d" % (N, M),
                         skipWarning=False)

import fipy.tools.dump
import fipy.tools.numerix
import fipy.tools.vector
from .dimensions.physicalField import PhysicalField
from fipy.tools.numerix import *
from fipy.tools.vitals import Vitals

__all__ = ["serialComm",
           "parallelComm",
           "dump",
           "numerix",
           "vector",
           "PhysicalField",
           "Vitals",
           "serial",
           "parallel"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

import os
if 'FIPY_INCLUDE_NUMERIX_ALL' in os.environ:
    import warnings
    class FiPyDeprecationWarning(DeprectationWarning):
        pass
    warnings.warn("""
The ability to include `numerix` functions in the `fipy` namespace
with the FIPY_INCLUDE_NUMERIX_ALL environment variable will
likely be removed in the future. If needed, the same effect can be
accomplished with `from fipy.tools.numerix import *`
""", FutureWarning)
    __all__.extend(numerix.__all__)



from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from builtins import object
from builtins import zip
from matplotlib import cm
from matplotlib import pyplot
from matplotlib import rcParams
from matplotlib import ticker
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import Normalize
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle

from scipy.io import mmio

from fipy.tools import numerix

__all__ = ["MatplotlibSparseMatrixViewer"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class MatplotlibSparseMatrixViewer(object):
    def __init__(self, title="Sparsity"):
        self.title = title

        self.L_width = 0.8
        self.margin = (1. - self.L_width) / 2
        self.b_width = self.margin
        self.c_width = self.margin / 3
        self.buffer = 1.5 * self.margin
        self.aspect = (self.margin + self.L_width                   # M
                       + self.buffer + self.c_width                 # colorbar
                       + self.buffer + self.b_width + self.margin)  # b

        pyplot.ion()

        fig = pyplot.figure(figsize=pyplot.figaspect(1. / self.aspect))
        self.id = fig.number

    def plot(self, matrix, RHSvector, log='auto'):
        import tempfile
        import os

        if "print" in os.environ['FIPY_DISPLAY_MATRIX'].lower().split():
            print("-"*75)
            print(self.title)
            print("-"*75)
            print("L:")
            print(matrix)
            print("b:", RHSvector)

        (f, mtxName) = tempfile.mkstemp(suffix='.mtx')
        matrix.exportMmf(mtxName)
        mtx = mmio.mmread(mtxName)
        os.remove(mtxName)

        pyplot.ion()

        c = mtx.tocoo()
        y = c.row
        x = c.col
        z = c.data
        N = matrix._shape[0]

        b = RHSvector
        if numerix.shape(b) == ():
            b = numerix.zeros((N,), 'l')

        if len(z) == 0:
            y = numerix.zeros((1,), 'l')
            x = numerix.zeros((1,), 'l')
            z = numerix.zeros((1,), 'l')

        def signed_to_logs(v):
            return (numerix.where(v > 0, numerix.log10(v), numerix.nan),
                    numerix.where(v < 0, numerix.log10(-v), numerix.nan))

        def logs_to_signed(v, plus, minus):
            v = numerix.where(v > 0, plus, -minus)
            v = numerix.where(numerix.isnan(v), 0., v)

            return v

        zPlus, zMinus = signed_to_logs(z)
        bPlus, bMinus = signed_to_logs(b)

        logs = (zPlus, zMinus, bPlus, bMinus)

        log = ((log == True)
               or (log == 'auto'
                   and (numerix.nanmax(numerix.concatenate(logs))
                        - numerix.nanmin(numerix.concatenate(logs)) > 2)))

        if log:
            zMin = numerix.nanmin(numerix.concatenate(logs))
            zMax = numerix.nanmax(numerix.concatenate(logs))

            zMin -= 0.5

            numdec = numerix.floor(zMax) - numerix.ceil(zMin)
            if numdec < 0:
                zMax += 0.5

            for v in logs:
                v -= zMin

            zRange = zMax - zMin

            if zRange == 0:
                zRange = numerix.nanmax(zPlus) + 1

            z = logs_to_signed(z, zPlus, zMinus)
            b = logs_to_signed(b, bPlus, bMinus)

            fmt = None
            loc = ticker.SymmetricalLogLocator(linthresh=1, base=10)

        else:
            zRange = max(abs(numerix.concatenate((z, b))))

            if zRange == 0:
                zRange = 1

            fmt = None
            loc = None


        pyplot.ioff()

        fig = pyplot.figure(self.id)
        fig.clf()

        usetex = rcParams['text.usetex']
        rcParams['text.usetex'] = False

        cmap = cm.RdBu

        norm = Normalize(vmin=-zRange, vmax=zRange)

        x0 = self.margin
        L_ax = fig.add_axes([x0 / self.aspect, self.margin, self.L_width / self.aspect, self.L_width])
        L_ax.text(0.5, -0.1, "L",
                  transform=L_ax.transAxes, horizontalalignment='center', verticalalignment='baseline')

        x0 += self.L_width + self.buffer
        c_ax = fig.add_axes([x0 / self.aspect, self.margin, self.c_width / self.aspect, self.L_width])

        x0 += self.c_width + self.buffer
        b_ax = fig.add_axes([x0 / self.aspect, self.margin, self.b_width / self.aspect, self.L_width],
                            sharey=L_ax)
        b_ax.text(0.5, -0.1, "b",
                  transform=b_ax.transAxes, horizontalalignment='center', verticalalignment='baseline')



        def scatterRectangles(x, y, z, norm=None, cmap=None):
            patches = [Rectangle(numerix.array([X - 0.5, Y - 0.5]), 1., 1.,
                                 edgecolor='none') for X, Y in zip(x, y)]

            collection = PatchCollection(patches, norm=norm, cmap=cmap,
                                         edgecolors='none')
            collection.set_array(z)

            return collection

        L_ax.add_collection(scatterRectangles(x=x, y=y, z=z,
                                              norm=norm, cmap=cmap))

        b_ax.add_collection(scatterRectangles(x=numerix.zeros((N,), 'l'), y=numerix.arange(N), z=b,
                                              norm=norm, cmap=cmap))

        ColorbarBase(ax=c_ax, cmap=cmap, norm=norm, orientation='vertical',
                     format=fmt, ticks=loc)

        pyplot.setp((b_ax.get_xticklabels(),
                     b_ax.get_yticklabels(),
                     b_ax.get_xticklines(),
                     b_ax.get_yticklines()), visible=False)

        L_ax.set_xlim(xmin=-0.5, xmax=N-0.5)
        L_ax.set_ylim(ymax=-0.5, ymin=N-0.5)

        b_ax.set_xlim(xmin=-0.5, xmax=0.5)
        b_ax.set_ylim(ymax=-0.5, ymin=N-0.5)

        fig.suptitle(self.title, x=0.5, y=0.95, fontsize=14)

        pyplot.draw()

        rcParams['text.usetex'] = usetex



from __future__ import unicode_literals
from builtins import zip
__docformat__ = 'restructuredtext'

from fipy.tools import numerix

from fipy.viewers.matplotlibViewer.matplotlibViewer import AbstractMatplotlibViewer

__all__ = ["Matplotlib2DViewer"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class AbstractMatplotlib2DViewer(AbstractMatplotlibViewer):
    def figaspect(self, figaspect, colorbar):
        if figaspect == 'auto':
            figaspect = self.vars[0].mesh.aspect2D
            # We can't make the colorbar, yet (as we don't even have a figure)
            # so hardcode these values, based on
            # https://matplotlib.org/api/_as_gen/matplotlib.figure.Figure.html#matplotlib.figure.Figure.colorbar
            #   fraction	0.15; fraction of original axes to use for colorbar
            #   pad	        0.05 if vertical, 0.15 if horizontal; fraction of original axes between colorbar and new image axes
            if colorbar == "vertical":
                figaspect = figaspect * (1. - (0.15 + 0.05))
            elif colorbar == "horizontal":
                figaspect = figaspect / (1. - (0.15 + 0.15))
        return figaspect

    def _plot(self):
        zmin, zmax = self._autoscale(vars=self.vars,
                                     datamin=self._getLimit(('datamin', 'zmin')),
                                     datamax=self._getLimit(('datamax', 'zmax')))

        self.norm.vmin = zmin
        self.norm.vmax = zmax

        self.mappable.set_norm(self.norm)

class Matplotlib2DViewer(AbstractMatplotlib2DViewer):
    """
    Displays a contour plot of a 2D `CellVariable` object.

    The `Matplotlib2DViewer` plots a 2D `CellVariable` using Matplotlib_.

    .. _Matplotlib: http://matplotlib.sourceforge.net/
    """

    __doc__ += AbstractMatplotlib2DViewer._test2Dirregular(viewer="Matplotlib2DViewer")

    def __init__(self, vars, title=None, limits={}, cmap=None, colorbar='vertical', axes=None, figaspect='auto', **kwlimits):
        """Creates a `Matplotlib2DViewer`.

        Parameters
        ----------
        vars : ~fipy.variables.cellVariable.CellVariable
            the `Variable` to display.
        title : str, optional
            displayed at the top of the `Viewer` window
        limits : dict, optional
            a (deprecated) alternative to limit keyword arguments
        cmap : ~matplotlib.colors.Colormap, optional
            the :class:`~matplotlib.colors.Colormap`.
            Defaults to `matplotlib.cm.jet`
        float xmin, xmax, ymin, ymax, datamin, datamax : float, optional
            displayed range of data. Any limit set to
            a (default) value of `None` will autoscale.
        colorbar : bool, optional
            plot a color bar in specified orientation if not `None`
        axes : ~matplotlib.axes.Axes, optional
            if not `None`, `vars` will be plotted into this Matplotlib `Axes` object
        figaspect : float, optional
            desired aspect ratio of figure. If arg is a number, use that aspect
            ratio. If arg is `auto`, the aspect ratio will be determined from
            the Variable's mesh.
        """
        kwlimits.update(limits)
        AbstractMatplotlib2DViewer.__init__(self, vars=vars, title=title, figaspect=figaspect,
                                            cmap=cmap, colorbar=colorbar, axes=axes,
                                            **kwlimits)

        self.mesh = self.vars[0].mesh

        vertexIDs = self.mesh._orderedCellVertexIDs

        vertexCoords = self.mesh.vertexCoords

        xCoords = numerix.take(vertexCoords[0], vertexIDs)
        yCoords = numerix.take(vertexCoords[1], vertexIDs)

        polys = []

        for x, y in zip(xCoords.swapaxes(0, 1), yCoords.swapaxes(0, 1)):
            if hasattr(x, 'mask'):
                x = x.compressed()
            if hasattr(y, 'mask'):
                y = y.compressed()
            polys.append(list(zip(x, y)))

        from matplotlib.collections import PolyCollection
        self.collection = PolyCollection(polys)
        self.collection.set_linewidth(0.5)
        try:
            self.axes.add_patch(self.collection)
        except:
            # PolyCollection not child of PatchCollection in matplotlib 0.98
            self.axes.add_collection(self.collection)

        xmin = self._getLimit('xmin', default=xCoords.min())
        xmax = self._getLimit('xmax', default=xCoords.max())
        ymin = self._getLimit('ymin', default=yCoords.min())
        ymax = self._getLimit('ymax', default=yCoords.max())

        self.axes.set_xlim(xmin=xmin, xmax=xmax)
        self.axes.set_ylim(ymin=ymin, ymax=ymax)

        self._plot()

    def _getSuitableVars(self, vars):
        from fipy.meshes.mesh2D import Mesh2D
        from fipy.variables.cellVariable import CellVariable
        vars = [var for var in AbstractMatplotlib2DViewer._getSuitableVars(self, vars) \
          if ((var.mesh.dim == 2 and isinstance(var, CellVariable))
              and var.rank == 0)]
        if len(vars) == 0:
            from fipy.viewers import MeshDimensionError
            raise MeshDimensionError("Matplotlib2DViewer can only display a rank-0, 2D CellVariable")
        # this viewer can only display one variable
        return [vars[0]]

    def _plot(self):
        super(Matplotlib2DViewer, self)._plot()

##         plt.clf()

##         ## Added garbage collection since matplotlib objects seem to hang
##         ## around and accumulate.
##         import gc
##         gc.collect()

        Z = self.vars[0].value

        rgba = self.cmap(self.norm(Z))

        self.collection.set_facecolors(rgba)
        self.collection.set_edgecolors(rgba)

##        plt.xlim(xmin=self._getLimit('xmin'),
##                 xmax=self._getLimit('xmax'))

##        plt.ylim(ymin=self._getLimit('ymin'),
##                 ymax=self._getLimit('ymax'))

if __name__ == "__main__":
    import fipy.tests.doctestPlus
    fipy.tests.doctestPlus.execButNoTest()



from __future__ import division
from __future__ import unicode_literals
__docformat__ = 'restructuredtext'

__all__ = ['SurfactantConvectionVariable']
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

from fipy.tools.numerix import MA
from fipy.tools import numerix

from fipy.tools import vector

from fipy.variables.faceVariable import FaceVariable

class SurfactantConvectionVariable(FaceVariable):
    """

    Convection coefficient for the `ConservativeSurfactantEquation`.
    The coefficient only has a value for a negative `distanceVar`.

    """

    def __init__(self, distanceVar):
        """

        Simple one dimensional test:


           >>> from fipy.variables.cellVariable import CellVariable
           >>> from fipy.meshes import Grid2D
           >>> mesh = Grid2D(nx = 3, ny = 1, dx = 1., dy = 1.)
           >>> from fipy.variables.distanceVariable import DistanceVariable
           >>> distanceVar = DistanceVariable(mesh, value = (-.5, .5, 1.5))
           >>> ## answer = numerix.zeros((2, mesh.numberOfFaces),'d')
           >>> answer = FaceVariable(mesh=mesh, rank=1, value=0.).globalValue
           >>> answer[0, 7] = -1
           >>> print(numerix.allclose(SurfactantConvectionVariable(distanceVar).globalValue, answer))
           True

        Change the dimensions:

           >>> mesh = Grid2D(nx = 3, ny = 1, dx = .5, dy = .25)
           >>> distanceVar = DistanceVariable(mesh, value = (-.25, .25, .75))
           >>> answer[0, 7] = -.5
           >>> print(numerix.allclose(SurfactantConvectionVariable(distanceVar).globalValue, answer))
           True

        Two dimensional example:

           >>> mesh = Grid2D(nx = 2, ny = 2, dx = 1., dy = 1.)
           >>> distanceVar = DistanceVariable(mesh, value = (-1.5, -.5, -.5, .5))
            >>> answer = FaceVariable(mesh=mesh, rank=1, value=0.).globalValue
           >>> answer[1, 2] = -.5
           >>> answer[1, 3] = -1
           >>> answer[0, 7] = -.5
           >>> answer[0, 10] = -1
           >>> print(numerix.allclose(SurfactantConvectionVariable(distanceVar).globalValue, answer))
           True

        Larger grid:

           >>> mesh = Grid2D(nx = 3, ny = 3, dx = 1., dy = 1.)
           >>> distanceVar = DistanceVariable(mesh, value = (1.5, .5, 1.5,
           ...                                           .5, -.5, .5,
           ...                                           1.5, .5, 1.5))
            >>> answer = FaceVariable(mesh=mesh, rank=1, value=0.).globalValue
           >>> answer[1, 4] = .25
           >>> answer[1, 7] = -.25
           >>> answer[0, 17] = .25
           >>> answer[0, 18] = -.25
           >>> print(numerix.allclose(SurfactantConvectionVariable(distanceVar).globalValue, answer))
           True

        """

        FaceVariable.__init__(self, mesh=distanceVar.mesh, name='surfactant convection', rank=1)
        self.distanceVar = self._requires(distanceVar)

    def _calcValue(self):

        Nfaces = self.mesh.numberOfFaces
        M = self.mesh._maxFacesPerCell
        dim = self.mesh.dim
        cellFaceIDs = self.mesh.cellFaceIDs

        faceNormalAreas = self.distanceVar._levelSetNormals * self.mesh._faceAreas

        cellFaceNormalAreas = numerix.array(MA.filled(numerix.take(faceNormalAreas, cellFaceIDs, axis=-1), 0))
        norms = numerix.array(MA.filled(MA.array(self.mesh._cellNormals), 0))

        alpha = numerix.dot(cellFaceNormalAreas, norms)
        alpha = numerix.where(alpha > 0, alpha, 0)

        alphasum = numerix.sum(alpha, axis=0)
        alphasum += (alphasum < 1e-100) * 1.0
        alpha = alpha / alphasum

        phi = numerix.repeat(self.distanceVar[numerix.newaxis, ...], M, axis=0)
        alpha = numerix.where(phi > 0., 0, alpha)

        volumes = numerix.array(self.mesh.cellVolumes)
        alpha = alpha * volumes * norms

        value = numerix.zeros((dim, Nfaces), 'd')

        vector._putAdd(value, cellFaceIDs, alpha, mask=MA.getmask(MA.array(cellFaceIDs)))

##         value = numerix.reshape(value, (dim, Nfaces, dim))

        return -value / self.mesh._faceAreas

def _test():
    import fipy.tests.doctestPlus
    return fipy.tests.doctestPlus.testmod()

if __name__ == "__main__":
    _test()





from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from builtins import range
__docformat__ = 'restructuredtext'

import os

from pysparse.direct import superlu

from fipy.solvers.pysparse.pysparseSolver import PysparseSolver
from fipy.tools import numerix

DEBUG = False

__all__ = ["LinearLUSolver"]
from future.utils import text_to_native_str
__all__ = [text_to_native_str(n) for n in __all__]

class LinearLUSolver(PysparseSolver):
    """

    The `LinearLUSolver` solves a linear system of equations using
    LU-factorization. This method solves systems with a general
    non-symmetric coefficient matrix using partial pivoting.

    The `LinearLUSolver` is a wrapper class for the the Pysparse_
    `superlu.factorize()` method.

    .. _Pysparse: http://pysparse.sourceforge.net

    """

    def __init__(self, tolerance=1e-10, iterations=10,
                       maxIterations=10, precon=None):
        """
        Creates a `LinearLUSolver`.

        Parameters
        ----------
        tolerance : float
            Required error tolerance.
        iterations : int
            Maximum number of iterative steps to perform.
        precon : ~fipy.solvers.pysparse.preconditioners.preconditioner.Preconditioner
            *ignored*
        """

        iterations = min(iterations, maxIterations)

        super(LinearLUSolver, self).__init__(tolerance = tolerance,
                                             iterations = iterations)

    def _solve_(self, L, x, b):
        diag = L.takeDiagonal()
        maxdiag = max(numerix.absolute(diag))

        L = L * (1 / maxdiag)
        b = b * (1 / maxdiag)

        LU = superlu.factorize(L.matrix.to_csr())

        if DEBUG:
            import sys
            print(L.matrix, file=sys.stderr)

        error0 = numerix.sqrt(numerix.sum((L * x - b)**2))

        for iteration in range(self.iterations):
            errorVector = L * x - b

            if (numerix.sqrt(numerix.sum(errorVector**2)) / error0)  <= self.tolerance:
                break

            xError = numerix.zeros(len(b), 'd')
            LU.solve(errorVector, xError)
            x[:] = x - xError

        if 'FIPY_VERBOSE_SOLVER' in os.environ:
            from fipy.tools.debug import PRINT
            PRINT('iterations: %d / %d' % (iteration+1, self.iterations))
            PRINT('residual:', numerix.sqrt(numerix.sum(errorVector**2)))
